"""
Copyright (c) 2016- 2025, Wiliot Ltd. All rights reserved.

Redistribution and use of the Software in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

  2. Redistributions in binary form, except as used in conjunction with
  Wiliot's Pixel in a product or a Software update for such product, must reproduce
  the above copyright notice, this list of conditions and the following disclaimer in
  the documentation and/or other materials provided with the distribution.

  3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
  may be used to endorse or promote products or services derived from this Software,
  without specific prior written permission.

  4. This Software, with or without modification, must only be used in conjunction
  with Wiliot's Pixel or with Wiliot's cloud service.

  5. If any Software is provided in binary form under this license, you must not
  do any of the following:
  (a) modify, adapt, translate, or create a derivative work of the Software; or
  (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
  discover the source code or non-literal aspects (such as the underlying structure,
  sequence, organization, ideas, or algorithms) of the Software.

  6. If you create a derivative work and/or improvement of any Software, you hereby
  irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
  royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
  right and license to reproduce, use, make, have made, import, distribute, sell,
  offer for sale, create derivative works of, modify, translate, publicly perform
  and display, and otherwise commercially exploit such derivative works and improvements
  (as applicable) in conjunction with Wiliot's products and services.

  7. You represent and warrant that you are not a resident of (and will not use the
  Software in) a country that the U.S. government has embargoed for use of the Software,
  nor are you named on the U.S. Treasury Department’s list of Specially Designated
  Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
  You must not transfer, export, re-export, import, re-import or divert the Software
  in violation of any export or re-export control laws and regulations (such as the
  United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
  and use restrictions, all as then in effect

THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
(SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
(A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
(B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
(C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""

'''
Eduard Tatievski
'''
import re
import sys
import threading

import matplotlib.pyplot as plt
from wiliot_core import PacketList
from wiliot_tools.utils.wiliot_gui.wiliot_gui import WiliotGui

from wiliot_testers.calibration_test.calibration_test_by_tbp import *
from wiliot_core.packet_data import packet_list, packet, tag_collection
from wiliot_tools.test_equipment.test_equipment import Attenuator, EquipmentError
from wiliot_testers.wiliot_tester_log import *
import os
import pathlib
import datetime
import csv
import pandas as pd
import collections


class Harvester(object):

    def __init__(self):
        # Getting all wanted values
        self.new_bb_list = None
        self.num_of_chambers = 1  # Number of chambers connected
        self.min_tags_answering = 1
        self.gui_values = self.init_gui()
        # Create log folder if not exist and build new name
        self.init_name_path()
        # Create log class
        self.my_logger = WiliotTesterLog(self.run_name_date)
        self.my_logger.set_logger(self.log_path, tester_name='Harvester')
        # Connect to GW and Attenuator if selected
        self.init_hw()
        # Build new power list according to gui inputs
        # self.power_range = self.build_range()
        self.power = re.split('(d)', self.gui_values['power'])[0]
        self.my_logger.results_logger.info('Starting scan on power: {}'.format(self.power))
        self.frequency_range = list(range(int(self.gui_values['start_freq']), int(self.gui_values['stop_freq']) + 1,
                                          int(self.gui_values['step'])))
        self.all_results = pd.DataFrame({'Adva': [], 'Frequency': [], 'TBP': [], 'BB': []})
        self.scan()

    def calibrate_bb(self):
        if self.gw_obj.connected:
            self.gw_obj.reset_buffer()
            self.gw_obj.start_continuous_listener()
        for bb_val in reversed(self.bb_list):
            self.gw_obj.config_gw(energy_pattern_val=self.gui_values['pattern'],
                                  time_profile_val=[self.gui_values['onTime'], self.gui_values['totalTime']],
                                  output_power_val='pos3dBm',
                                  beacons_backoff_val=bb_val)

    def build_range(self):
        if not self.sub1g_enable:
            if self.attn_enable:
                for index, nrg in enumerate(self.abs_att_power_list):
                    if self.gui_values['lowerPower'] == nrg:
                        low_limit = index
                    if self.gui_values['higherPower'] == nrg:
                        high_limit = index
                power_range = self.abs_att_power_list[low_limit:high_limit + 1]
            else:
                for index, nrg in enumerate(self.abs_power_list):
                    if self.gui_values['lowerPower'] == nrg:
                        low_limit = index
                    if self.gui_values['higherPower'] == nrg:
                        high_limit = index
                power_range = self.abs_power_list[low_limit:high_limit + 1]

        else:
            if self.attn_enable:
                for index, nrg in enumerate(self.valid_attn_sub1g_output_power):
                    if self.gui_values['lowerPower'] == nrg:
                        low_limit = index
                    if self.gui_values['higherPower'] == nrg:
                        high_limit = index
                power_range = self.valid_attn_sub1g_output_power[low_limit:high_limit + 1]
            else:
                for index, nrg in enumerate(self.valid_sub1g_output_power):
                    if self.gui_values['lowerPower'] == nrg:
                        low_limit = index
                    if self.gui_values['higherPower'] == nrg:
                        high_limit = index
                power_range = self.valid_sub1g_output_power[low_limit:high_limit + 1]

        output_power = []
        for power in power_range:
            if int(power[:-3]) < 0:
                output_power.append('neg' + power[1:])
            else:
                output_power.append('pos' + power)

        return output_power

    def convert_power(self, power, sub1g=False):
        gw_power = None
        attn_power = None
        # Convert from neg3dBm to -3
        if power[:3] == 'neg':
            abs_power = 0 - int(power[3:-3])
        elif power[:3] == 'pos':
            abs_power = int(power[3:-3])
        else:
            abs_power = int(power[:-3])

        # Find the value for attenuator if available
        if sub1g:
            if self.attn_enable:
                for power_scan in self.valid_sub1g_output_power:
                    if int(power_scan[:-3]) > abs_power:
                        gw_power = int(power_scan[:-3])
                        attn_power = int(gw_power[:-3]) - abs_power
            else:
                gw_power = abs_power
                attn_power = 0

        else:
            if self.attn_enable:
                for power_scan in self.abs_power_list:
                    if int(power_scan[:-3]) > abs_power:
                        gw_power = int(power_scan[:-3])
                        attn_power = int(gw_power[:-3]) - abs_power
            else:
                gw_power = abs_power
                attn_power = 0
        return gw_power, attn_power

    def init_hw(self):

        try:
            self.gw_obj = WiliotGateway(auto_connect=True, logger_name=self.my_logger.gw_logger.name, verbose=False)
            time.sleep(1)
            if self.gw_obj.connected:
                self.gw_obj.reset_gw()
                self.gw_obj.reset_buffer()
                time.sleep(1.5)
                self.gw_obj.start_continuous_listener()
                self.gw_obj.write('!set_tester_mode 1', with_ack=True)
                self.gw_obj.write('!pl_gw_config 1', with_ack=True)
                self.gw_obj.write('!enable_brg_mgmt 0', with_ack=True)
                self.gw_obj.write('!listen_to_tag_only 1', with_ack=True)
                self.gw_obj.write('!energizing_prob_set 100', with_ack=True)
                self.my_logger.gw_logger.info('GW config finished')
            else:
                raise EquipmentError('Gateway Error - Verify WiliotGateway connection')
        except Exception as e:
            self.my_logger.gw_logger.warning(e)
            raise EquipmentError('Gateway Error - Verify WiliotGateway connection')
            self.gw_obj.close_port()

        if self.attn_enable:
            self.attntype = self.gui_values['attntype']
            self.attnval = int(self.gui_values['attnval'])

            try:
                self.attn_obj = Attenuator(str(self.attntype)).GetActiveTE()
                current_attn = self.attn_obj.Getattn()
                self.my_logger.gw_logger.info('Current attenuator values: {}'.format(current_attn))
            except Exception:
                raise EquipmentError('Attenuator Error - Verify Attenuator connection')

    def init_name_path(self):
        self.log_path = os.path.join('', 'logs')
        if not os.path.exists(self.log_path):
            pathlib.Path(self.log_path).mkdir(parents=True, exist_ok=True)
        self.time_start = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
        self.run_name_date = str(self.gui_values['testName']) + '_' + str(self.time_start)
        self.full_name_path = os.path.join(self.log_path, self.run_name_date)
        self.packets_csv_path = os.path.join(self.full_name_path, self.run_name_date + '.csv')
        self.test_csv_path = os.path.join(self.full_name_path, self.run_name_date + '_tests.csv')

    def init_gui(self):

        # self.bb_list = [0, 2, 7, 12, 19, 20, 21, 22, 23, 24, 25, 27, 29, 30, 31, 32, 33, 36, 40]
        self.bb_list = [40, 36, 33, 32, 31, 30, 29, 27, 25, 24, 23, 22, 21, 20, 19, 12, 7, 2, 0]
        # self.bb_list = [12, 7, 2, 0]

        self.valid_sub1g_output_power = ['17dBm', '18dBm', '19dBm', '20dBm', '21dBm', '22dBm',
                                         '23dBm', '24dBm', '25dBm', '26dBm', '27dBm', '28dBm', '29dBm']

        self.valid_attn_sub1g_output_power = ['-10dBm', '-9dBm', '-8dBm', '-7dBm', '-6dBm', '-5dBm',
                                              '-4dBm', '-3dBm', '-2dBm', '-1dBm', '0dBm', '1dBm',
                                              '2dBm', '3dBm', '4dBm', '5dBm', '6dBm', '7dBm', '8dBm',
                                              '9dBm', '10dBm', '11dBm', '12dBm', '13dBm', '14dBm', '15dBm',
                                              '16dBm', '17dBm', '18dBm', '19dBm', '20dBm', '21dBm', '22dBm',
                                              '23dBm', '24dBm']

        self.abs_power_list = ['-21dBm', '-18dBm', '-15dBm', '-12dBm', '-8dBm', '-5dBm', '-2dBm',
                               '-1dBm', '0dBm', '1dBm', '2dBm', '3dBm', '4dBm', '6dBm',
                               '10dBm', '15dBm', '20dBm', '21dBm', '22dBm']

        self.abs_att_power_list = ['-35dbm', '-34dbm', '-33dbm', '-32dbm', '-31dbm', '-30dbm',
                                   '-29dbm', '-28dbm', '-27dbm', '-26dbm', '-25dbm', '-24dbm',
                                   '-23dbm', '-22dbm', '-21dbm', '-20dbm', '-19dbm', '-18dbm',
                                   '-17dbm', '-16dbm', '-15dbm', '-14dbm', '-13dbm', '-12dbm',
                                   '-11dbm', '-10dbm', '-9dbm', '-8dbm', '-7dbm', '-6dbm',
                                   '-5dbm', '-4dbm', '-3dbm', '-2dbm', '-1dbm', '0dbm',
                                   '1dbm', '2dbm', '3dbm', '4dbm', '5dbm', '6dbm',
                                   '7dbm', '8dbm', '9dbm', '10dbm', '11dbm',
                                   '12dbm', '13dbm', '14dbm', '15dbm', '16dbm', '17dbm']

        self.valid_output_power_vals = [
            {'abs_power': -21, 'gw_output_power': 'neg20dBm', 'bypass_pa': '1'},
            {'abs_power': -18, 'gw_output_power': 'neg16dBm', 'bypass_pa': '1'},
            {'abs_power': -15, 'gw_output_power': 'neg12dBm', 'bypass_pa': '1'},
            {'abs_power': -12, 'gw_output_power': 'neg8dBm', 'bypass_pa': '1'},
            {'abs_power': -8, 'gw_output_power': 'neg4dBm', 'bypass_pa': '1'},
            {'abs_power': -5, 'gw_output_power': 'pos0dBm', 'bypass_pa': '1'},
            {'abs_power': -2, 'gw_output_power': 'pos2dBm', 'bypass_pa': '1'},
            {'abs_power': -1, 'gw_output_power': 'pos3dBm', 'bypass_pa': '1'},
            {'abs_power': 0, 'gw_output_power': 'pos4dBm', 'bypass_pa': '1'},
            {'abs_power': 1, 'gw_output_power': 'pos5dBm', 'bypass_pa': '1'},
            {'abs_power': 2, 'gw_output_power': 'pos6dBm', 'bypass_pa': '1'},
            {'abs_power': 3, 'gw_output_power': 'pos7dBm', 'bypass_pa': '1'},
            {'abs_power': 4, 'gw_output_power': 'pos8dBm', 'bypass_pa': '1'},
            {'abs_power': 6, 'gw_output_power': 'neg12dBm', 'bypass_pa': '0'},
            {'abs_power': 10, 'gw_output_power': 'neg8dBm', 'bypass_pa': '0'},
            {'abs_power': 15, 'gw_output_power': 'neg4dBm', 'bypass_pa': '0'},
            {'abs_power': 20, 'gw_output_power': 'pos0dBm', 'bypass_pa': '0'},
            {'abs_power': 21, 'gw_output_power': 'pos2dBm', 'bypass_pa': '0'},
            {'abs_power': 22, 'gw_output_power': 'pos3dBm', 'bypass_pa': '0'}]

        params_dict = {
            'testName': {
                'text': 'Test Name',
                'value': '',
                'widget_type': 'entry',
            },
            'tagsNum': {
                'text': 'Num of tags',
                'value': self.min_tags_answering,
                'widget_type': 'entry',
            },
            'testTime': {
                'text': 'Test Time',
                'value': '10',
                'widget_type': 'entry',
            },
            'start_freq': {
                'text': 'Start [MHz]',
                'value': '2402',
                'widget_type': 'entry',
                'group': 'Frequency range'
            },
            'stop_freq': {
                'text': 'Stop [MHz]',
                'value': '2480',
                'widget_type': 'entry',
                'group': 'Frequency range'
            },
            'step': {
                'text': 'Step [MHz]',
                'value': '2',
                'widget_type': 'entry',
                'group': 'Frequency range'
            },
            'power': {
                'text': 'Energy',
                'value': self.abs_power_list[0],
                'options': self.abs_power_list,
                'widget_type': 'combobox',
            },
            'lowerBB': {
                'text': 'Beacons backoff (lower)',
                'value': self.bb_list[0],
                'options': self.bb_list,
                'widget_type': 'combobox',
            },
            'higherBB': {
                'text': 'Beacons backoff (higher)',
                'value': self.bb_list[-1],
                'options': self.bb_list,
                'widget_type': 'combobox',
            },
            '2.4mode': {
                'text': '2.4Ghz Mode',
                'value': '2.4Ghz',
                'widget_type': 'button',
            },
            'sub1gmode': {
                'text': 'Sub1G Mode',
                'value': 'Sub1G',
                'widget_type': 'button',
            },
            'onTime': {
                'text': 'Time profile - On',
                'value': '5',
                'widget_type': 'entry',
            },
            'totalTime': {
                'text': 'Off',
                'value': '15',
                'widget_type': 'entry',
            },
            'attn': {
                'text': 'Enable Attenuator',
                'value': 'Enable Attenuator',
                'widget_type': 'button',
            },
            '-ATTN-': {
                'text': 'Attenuator Disabled',
                'value': 'Attenuator Disabled',
                'widget_type': 'label',
                'options': {'fg': 'Dark Red'}
            },
            'attntype': {
                'text': 'Attenuator type',
                'value': 'API',
                'options': ('API', 'MCDI', 'MCDI-USB'),
                'widget_type': 'combobox',
            },
            'attntypeen': {
                'text': 'Attenuator type: ',
                'value': '',
                'widget_type': 'label',
            },
            'attnval': {
                'text': 'Attenuation in dBm',
                'value': '0',
                'widget_type': 'entry',
            },
            'attnvalen': {
                'text': 'Attenuation in dBm',
                'value': '',
                'widget_type': 'label',
            },
            '-OUTPUT-': {
                'text': '',
                'value': '',
                'widget_type': 'label',
                'options': {'fg': 'red', 'font': 'bold'},
            }
        }

        wg = WiliotGui(params_dict=params_dict, do_button_config=False, title='Harvest test')
        self.attn_enable = False
        self.sub1g_enable = False

        def on_submit_button():
            test_name = wg.widgets.get('testName').get()
            start_freq = int(wg.widgets.get('start_freq').get())
            stop_freq = int(wg.widgets.get('stop_freq').get())
            if ' ' in test_name or '/' in test_name or '\\' in test_name or test_name == '':
                wg.update_widget(widget_key='-OUTPUT-', new_value='Please enter valid run name without / or spaces',
                                 color='red')
            if self.sub1g_enable:
                if start_freq not in range(849, 941) or stop_freq not in range(849, 941):
                    wg.update_widget(widget_key='-OUTPUT-', new_value='Please enter valid Frequency range for Sub1G',
                                     color='red')
            else:
                if start_freq not in range(2400, 2490) or stop_freq not in range(2400, 2490):
                    wg.update_widget(widget_key='-OUTPUT-', new_value='Please enter valid Frequency range for 2.4GHz',
                                     color='red')
            self.min_tags_answering = int(wg.widgets.get('tagsNum').get())
            start_index = int(wg.widgets.get('lowerBB').get())
            higher_bb = int(wg.widgets.get('higherBB').get())
            end_index = self.bb_list.index(higher_bb)
            direction = -1 if start_index > end_index else 1
            self.new_bb_list = self.bb_list[start_index:end_index + direction:direction]
            wg.on_submit()

        def on_sub1gmode_click():

            self.sub1g_enable = True

            wg.update_widget(widget_key='2.4mode', new_value='2.4Ghz', color='#821515')
            wg.update_widget(widget_key='sub1gmode', new_value='Sub1G', color='#158225')
            wg.update_widget(widget_key='power', new_value=self.valid_sub1g_output_power[0], options=tuple(self.valid_sub1g_output_power))
            wg.update_widget(widget_key='start_freq', new_value='850')
            wg.update_widget(widget_key='stop_freq', new_value='940')

        def on_24mode_click():

            self.sub1g_enable = False

            wg.update_widget(widget_key='2.4mode', new_value='2.4Ghz', color='#158225')
            wg.update_widget(widget_key='sub1gmode', new_value='Sub1G', color='#821515')
            wg.update_widget(widget_key='power', new_value=self.abs_power_list[0], options=tuple(self.abs_power_list))

            # Update the 'start_freq' and 'stop_freq' entries
            wg.update_widget(widget_key='start_freq', new_value='2402')
            wg.update_widget(widget_key='stop_freq', new_value='2480')

        def on_attn_click():
            self.attn_enable = not self.attn_enable

            if self.attn_enable:
                wg.update_widget(widget_key='attn', new_value='Disable Attenuator', color='#821515')
                wg.update_widget(widget_key='-ATTN-', new_value='Attenuator Enabled', color='Dark Green')

                wg.update_widget(widget_key='attntype', disabled=False)
                wg.update_widget(widget_key='attntypeen', disabled=False)
                wg.update_widget(widget_key='attnval', disabled=False)
                wg.update_widget(widget_key='attnvalen', disabled=False)
            else:
                wg.update_widget(widget_key='attn', new_value='Enable Attenuator', color='#158225')
                wg.update_widget(widget_key='-ATTN-', new_value='Attenuator Disabled', color='Dark Red')

                wg.update_widget(widget_key='attntype', disabled=True)
                wg.update_widget(widget_key='attntypeen', disabled=True)
                wg.update_widget(widget_key='attnval', disabled=True)
                wg.update_widget(widget_key='attnvalen', disabled=True)

        wg.button_configs(submit_command=on_submit_button)
        wg.add_event(widget_key='sub1gmode', command=on_sub1gmode_click, event_type='button')
        wg.add_event(widget_key='2.4mode', command=on_24mode_click, event_type='button')
        wg.add_event(widget_key='attn', command=on_attn_click, event_type='button')
        on_attn_click()
        values = wg.run()
        return values

    def packet_filter(self, packet_list, filter_param='rssi', param_limits=None):
        """
        filter the packets from the packet_list according to the filter_param and the param_limits.
        packets that were filtered out of the test does no take into account during the test and considered as noise
        :param packet_list: the incoming packets list
        :type packet_list: PacketList
        :param filter_param: one of the packet dataframe column names
        :type filter_param: str
        :param param_limits: list of two elements [x, y] that defines the range of the valid param (x,y including)
        :type param_limits: list
        :return: the filtered packet list
        :rtype: PacketList
        """

        def update_custom_data(packet, key, value):
            if key in packet.custom_data.keys():
                packet.custom_data[key].append(value)
            else:
                packet.custom_data[key] = [value]

        if param_limits is None:
            param_limits = [0, float('inf')]
        filtered_packet_list = PacketList()

        for p in packet_list:
            # filter packets from black list:
            if p.packet_data['adv_address'] in self.black_list:
                p.add_custom_data(custom_data={'packet_status': 'blacklist'})
            else:
                # filter packets by filter param:
                packet_param_list = p.extract_packet_data_by_name(filter_param)
                sprinkler_ids = []
                for i, param in enumerate(packet_param_list):
                    if param_limits[0] <= param <= param_limits[-1]:
                        update_custom_data(p, 'packet_status', 'good')
                        sprinkler_ids.append(i)
                    else:
                        update_custom_data(p, 'packet_status', filter_param)

                if len(sprinkler_ids) > 0:
                    filtered_packet_list.append(p.filter_by_sprinkler_id(sprinkler_ids=sprinkler_ids))

        return filtered_packet_list

    def init_timer(self):
        self.timeout = False
        self.success = False
        self.clean_gw()
        self.timer = threading.Timer(int(self.gui_values['testTime']), self.time_expire)

    def start_analyze(self, power=None, freq=None, bb=None):
        tbp = None
        for tag_adva in self.all_tags.tags:
            if self.is_gw_fetal_error(self.all_tags.tags[tag_adva]):
                self.my_logger.gw_logger.warning('gw fetal error was detected')
                self.gw_reset_and_config()
            for packet in self.all_tags.tags[tag_adva].packet_list:
                self.all_tests.append(packet)
            try:
                tbp = self.all_tags.get_avg_tbp_by_id(tag_adva)
                packets_sum = self.all_tags.get_statistics_by_id(tag_adva)['num_packets']
            except Exception as e:
                self.my_logger.logger.info(
                    'Exception receiving TBP or Packets num on tag {}, exception {}'.format(tag_adva, e))
                tbp = None
            new_row = {'Adva': tag_adva, 'Frequency': freq, 'TBP': tbp, 'BB': bb}
            new_df = pd.DataFrame(new_row, index=[0])
            self.all_results = pd.concat([self.all_results, new_df], ignore_index=True)
            details = {'Test Num': self.test_num, 'BB': bb, 'Frequency': freq, 'Power': power, 'Packets': packets_sum,
                       'TBP': tbp}
            if self.all_tests.tags[tag_adva].list_custom_data.__len__() == 0:
                self.all_tests.tags[tag_adva].list_custom_data['Data'] = []
            self.all_tests.tags[tag_adva].list_custom_data['Data'].append(details)

    def get_packets(self, filter_param='rssi', filter_value=100, power=None, freq=None, bb=None):
        """
        receive packets from the gw (via the listener) and filter them according to packet_filter()
        :param filter_param: one of the packet dataframe column names (for packet_filter)
        :type filter_param: str
        :param filter_value: the max valid value of the specified param
        :type filter_value: int
        :return: True if packet were received
        :rtype: bool
        """
        packets_received_list = None
        self.num_good_tags = 0
        self.num_bad_tags = 0
        if self.gw_obj.is_data_available():
            packets_received_list = self.gw_obj.get_packets(action_type=ActionType.ALL_SAMPLE,
                                                            data_type=DataType.PACKET_LIST)
        if packets_received_list:
            for packet in packets_received_list:
                packet_param_list = int(packet.get_average_rssi())
                if filter_value >= packet_param_list:
                    self.my_logger.logger.info('got packet {}'.format(packet.packet_data['raw_packet']))
                    self.all_tags.append(packet=packet.copy())
                else:
                    self.my_logger.logger.info(
                        'got packet {} but with high rssi {}'.format(packet.packet_data['raw_packet'],
                                                                     packet_param_list))
                    continue

                if self.all_tags.tags.__len__() > 0:
                    for packet_list in self.all_tags.tags:
                        try:
                            avg_tbp = self.all_tags.tags[packet_list].get_avg_tbp()
                            if avg_tbp > 0:
                                self.num_good_tags += 1
                        except Exception:
                            pass
                        if self.num_good_tags >= self.min_tags_answering:
                            self.num_bad_tags = self.all_tags.tags.__len__() - self.num_good_tags
                            self.success = True
                        else:
                            continue
                if self.success:
                    self.my_logger.logger.info(
                        'Found {} tags\n{} tags TBP was calculated\n{} tags TBP wasnt calculated\nStarting analyzing'.format(
                            self.all_tags.tags.__len__(), self.num_good_tags, self.num_bad_tags))
                    break

        if self.success:
            self.timer.cancel()
            self.gw_obj.stop_gw_app()
            self.clean_gw()
            self.gw_obj.stop_continuous_listener()
            self.start_analyze(power, freq, bb)

    def is_gw_fetal_error(self, packet_list):
        """
        check if the gw reset itself for some reason
        :param packet_list: the received packets list
        :type packet_list: PacketList
        :return:
        :rtype:
        """
        counter = 0
        for p in packet_list:
            if p.gw_data['stat_param'].size > 1:
                for s in p.gw_data['stat_param']:
                    if s == 0:
                        counter = counter + 1
                        if counter > 1:
                            return True
            else:
                if p.gw_data['stat_param'] == 0:
                    counter = counter + 1
                    if counter > 1:
                        return True
        return False

    def get_result(self):
        return self.all_tests

    def time_expire(self):
        self.timeout = True
        self.timer.cancel()
        # self.gw_obj.stop_gw_app()
        self.my_logger.results_logger.info(
            'Timeout - found {} tags but didnt received enough packets to calculate TBP for all, calculated only for {} tags'.format(
                self.all_tags.tags.__len__(), self.num_good_tags))

    def gw_reset_and_config(self):
        self.gw_obj.reset_gw()
        time.sleep(2)
        self.gw_obj.write('!set_tester_mode 1', with_ack=True)
        self.gw_obj.write('!enable_brg_mgmt 0', with_ack=True)
        self.gw_obj.write('!listen_to_tag_only 1', with_ack=True)
        self.gw_obj.write('!energizing_prob_set 100', with_ack=True)

    def clean_gw(self):
        self.gw_obj.reset_buffer()
        self.gw_obj.reset_listener()

    def data_read_analyze(self, dict=None):

        # Get unique BB values
        self.all_results = self.all_results.dropna(subset=['TBP'])

        unique_bbs = self.all_results['BB'].unique()

        # Set up subplots
        fig, axes = plt.subplots(nrows=1, ncols=unique_bbs.size, figsize=(10, 4))

        # Loop over unique BB values and plot graphs
        for i, bb in enumerate(unique_bbs):
            # Get data for current BB value
            bb_df = self.all_results[self.all_results['BB'] == bb]
            # Get unique Adva names for current BB value
            unique_adv = bb_df['Adva'].unique()

            # Plot graph for current BB value on the appropriate subplot
            ax = axes[i]
            ax.set_title(f"Beacon Backoff = {bb}")
            freq_ticks = sorted(self.all_results['Frequency'].unique())
            ax.set_xticks(self.x_grid, rotation=90)
            ax.set_xlabel("Frequency")
            ax.set_ylabel("TBP")
            for adv in unique_adv:
                adv_df = bb_df[bb_df['Adva'] == adv]
                ax.plot(adv_df['Frequency'], adv_df['TBP'], label=adv)
            ax.legend()

        # Show the plot
        plt.show()

    def finish(self):
        # Start writing all data to file
        self.all_results = self.all_results.groupby('Adva').apply(
            lambda x: x.sort_values(by=['Frequency', 'BB'], ascending=[True, False])).reset_index(drop=True)

        self.my_logger.results_logger.info(self.all_results)
        for i in self.all_results:
            self.my_logger.results_logger.info(i)
        self.all_packet_list = self.all_tests.to_packet_list()
        all_data = []
        custom_data = {'Test Num': '', 'Adva': '', 'BeaconBackoff': '', 'Frequency': '', 'Power': '', 'Packets': '',
                       'TBP': ''}
        for packet_list in self.all_tests.tags:
            for custom in self.all_tests.tags[packet_list].list_custom_data['Data']:
                p_list_data = {'Test Num': '', 'Adva': '', 'BeaconBackoff': '', 'Frequency': '', 'Power': '',
                               'Packets': '', 'TBP': ''}
                p_list_data['Adva'] = packet_list
                p_list_data['Test Num'] = custom['Test Num']
                p_list_data['BeaconBackoff'] = custom['BB']
                p_list_data['Frequency'] = custom['Frequency']
                p_list_data['Power'] = custom['Power']
                p_list_data['Packets'] = custom['Packets']
                p_list_data['TBP'] = custom['TBP'] if custom['TBP'] is not None else 0
                all_data.append(p_list_data)
        try:
            with open(self.test_csv_path, 'w', newline='') as csvfile:
                writer = csv.DictWriter(csvfile, fieldnames=custom_data)
                writer.writeheader()
                writer.writerows(all_data)

        except IOError:
            print("I/O error")

        try:
            self.all_tests.to_csv(path=self.packets_csv_path, val='adv_address')
        except Exception:
            self.my_logger.logger.warning('Empty list')
        # Read data from file organized with pandas

        # Draw graphs according to post process correlated to TBP requested
        if len(self.frequency_range) > 8:
            self.x_grid = self.frequency_range[::3]
            if len(self.frequency_range) > 15:
                self.x_grid = self.frequency_range[::4]
                if len(self.frequency_range) > 28:
                    self.x_grid = self.frequency_range[::5]
        else:
            self.x_grid = self.frequency_range
        self.data_read_analyze()
        # The End
        self.my_logger.logger.info('Finished')

    def config_gw(self, power, freq, bb):
        if not self.sub1g_enable:
            # ---------------------------------------------------------------------------------------------------------------------
            # 2.4 + Attenuator
            if self.attn_enable:
                # abs_power_val, abs_attn_val = self.convert_power(power, sub1g=False)
                # for list_power_val in self.valid_output_power_vals:
                #     if abs_power_val == list_power_val['abs_power']:
                #         gw_out_power = list_power_val['gw_output_power']
                self.my_logger.logger.info(
                    'Setting GW power: {}, Attenuation value: {} Frequency {} Beacaon Backoff val {}'.format(
                        power, self.attnval, freq, bb))
                self.attn_obj.Setattn(self.attnval)
                self.gw_obj.config_gw(energy_pattern_val=18,
                                      time_profile_val=[self.gui_values['onTime'],
                                                        self.gui_values['totalTime']],
                                      beacons_backoff_val=bb,
                                      received_channel='37',
                                      effective_output_power_val=int(power),
                                      start_gw_app=True,
                                      with_ack=True)
            # ---------------------------------------------------------------------------------------------------------------------
            # 2.4 + No Attenuator
            else:
                # abs_power_val, abs_attn_val = self.convert_power(power, sub1g=False)
                # if int(abs_power_val) > 5:
                #     bypass_pa_val = 0
                # else:
                #     bypass_pa_val = 1
                # for list_power_val in self.valid_output_power_vals:
                #     if abs_power_val == int(list_power_val['abs_power']):
                #         gw_out_power = list_power_val['gw_output_power']
                self.my_logger.logger.info(
                    'Setting GW power: {}, No attenuation Frequency {} Beacaon Backoff val {}'.format(
                        power, freq, bb))
                self.gw_obj.config_gw(energy_pattern_val=18,
                                      time_profile_val=[self.gui_values['onTime'],
                                                        self.gui_values['totalTime']],
                                      beacons_backoff_val=bb,
                                      received_channel='37',
                                      effective_output_power_val=int(power),
                                      start_gw_app=True,
                                      with_ack=True)
        # ---------------------------------------------------------------------------------------------------------------------
        # Sub1G
        else:
            # ---------------------------------------------------------------------------------------------------------------------
            # Sub1G + Attenuator
            if self.attn_enable:
                # abs_power_val, abs_attn_val = self.convert_power(power, sub1g=True)
                self.my_logger.logger.info(
                    'Setting Sub1G GW power: {}, Attenuation value: {} Frequency {} Beacaon Backoff val {}'.format(
                        power, self.attnval, freq, bb))
                self.attn_obj.Setattn(self.attnval)
                self.gw_obj.config_gw(energy_pattern_val=50,
                                      time_profile_val=[self.gui_values['onTime'],
                                                        self.gui_values['totalTime']],
                                      beacons_backoff_val=bb,
                                      received_channel='37',
                                      effective_output_power_val=22,
                                      sub1g_output_power_val=power,
                                      start_gw_app=True,
                                      with_ack=True)
            # ---------------------------------------------------------------------------------------------------------------------
            # Sub1G + No Attenuator
            else:
                # abs_power_val, abs_attn_val = self.convert_power(power, sub1g=True)
                self.my_logger.logger.info(
                    'Setting Sub1G GW power: {}, No attenuation value Frequency {} Beacaon Backoff val {}'.format(
                        power, freq, bb))
                self.gw_obj.config_gw(energy_pattern_val=50,
                                      time_profile_val=[self.gui_values['onTime'],
                                                        self.gui_values['totalTime']],
                                      beacons_backoff_val=bb,
                                      received_channel='37',
                                      effective_output_power_val=22,
                                      sub1g_output_power_val=power,
                                      start_gw_app=True,
                                      with_ack=True)
        # ---------------------------------------------------------------------------------------------------------------------
        if self.sub1g_enable:
            self.gw_obj.write('!set_dyn_energizing_pattern 6 1 0 0', with_ack=True)
            self.gw_obj.write('!set_beacons_pattern 550 3 2402 2426 2480', with_ack=True)
            self.gw_obj.write('!set_sub_1_ghz_energizing_frequency {}'.format(freq), with_ack=True)

        else:
            self.gw_obj.write('!set_dyn_energizing_pattern 6 0 1 {}'.format(freq), with_ack=True)
            self.gw_obj.write('!set_beacons_pattern 550 3 2402 2426 2480', with_ack=True)

    def scan(self):
        self.all_tests = TagCollection()
        self.test_num = 0
        self.results = []
        # Scan power
        power = self.power
        self.my_logger.logger.info('Sweeping freq {}'.format(self.frequency_range))
        # Scan frequency
        for freq in self.frequency_range:
            # Scan beacon backoff
            for index, bb in enumerate(self.new_bb_list):
                self.test_num += 1
                # 2.4Ghz
                self.all_tags = TagCollection()
                self.gw_obj.start_continuous_listener()
                self.init_timer()  # Init timeout flag and reset success flag
                self.timer.start()  # Start timer thread
                self.my_logger.results_logger.info(
                    '-' * 30 + 'New Test - Power {} Frequency {} Beacon Backoff {}'.format(power, freq,
                                                                                           bb) + '-' * 30)

                self.config_gw(power=power, freq=freq, bb=bb)

                # Start receiving packets
                while not self.timeout and not self.success:  # While function to work until timeout or success test
                    time.sleep(0.01)
                    self.get_packets(power=power, freq=freq,
                                     bb=bb)  # Will start to get packets, then analyze them and add to self.all_tests with custom data of which test it came from

                self.clean_gw()  # clear the buffers
                time.sleep(0.5)

        self.gw_obj.close_port(is_reset=True)
        self.gw_obj.stop_continuous_listener()
        self.finish()


if __name__ == '__main__':
    harvest = Harvester()
    results = harvest.get_result()
    pass



"""
  Copyright (c) 2016- 2025, Wiliot Ltd. All rights reserved.

  Redistribution and use of the Software in source and binary forms, with or without modification,
   are permitted provided that the following conditions are met:

     1. Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

     2. Redistributions in binary form, except as used in conjunction with
     Wiliot's Pixel in a product or a Software update for such product, must reproduce
     the above copyright notice, this list of conditions and the following disclaimer in
     the documentation and/or other materials provided with the distribution.

     3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
     may be used to endorse or promote products or services derived from this Software,
     without specific prior written permission.

     4. This Software, with or without modification, must only be used in conjunction
     with Wiliot's Pixel or with Wiliot's cloud service.

     5. If any Software is provided in binary form under this license, you must not
     do any of the following:
     (a) modify, adapt, translate, or create a derivative work of the Software; or
     (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
     discover the source code or non-literal aspects (such as the underlying structure,
     sequence, organization, ideas, or algorithms) of the Software.

     6. If you create a derivative work and/or improvement of any Software, you hereby
     irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
     royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
     right and license to reproduce, use, make, have made, import, distribute, sell,
     offer for sale, create derivative works of, modify, translate, publicly perform
     and display, and otherwise commercially exploit such derivative works and improvements
     (as applicable) in conjunction with Wiliot's products and services.

     7. You represent and warrant that you are not a resident of (and will not use the
     Software in) a country that the U.S. government has embargoed for use of the Software,
     nor are you named on the U.S. Treasury Department’s list of Specially Designated
     Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
     You must not transfer, export, re-export, import, re-import or divert the Software
     in violation of any export or re-export control laws and regulations (such as the
     United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
     and use restrictions, all as then in effect

   THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
   OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
   WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
   QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
   IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
   ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
   OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
   FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
   (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
   (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
   (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
   (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""
import time
import datetime
import pandas as pd
import os
import multiprocessing

from wiliot_core import set_logger, QueueHandler
from wiliot_tools.test_equipment.test_equipment import CognexNetwork, CognexDataMan, CognexDebug
from wiliot_tools.association_tool.association_configs import is_wiliot_code
from wiliot_tools.association_tool.association_tool import CloudAssociation


MACHINE_PING_TIMEOUT = 10 # seconds, if no codes were scanned for this time, the scanner will be triggered again


class AssociatorProcess(CloudAssociation):
    def __init__(self, associate_q=None, stop_event=None, owner_id='', category_id='', time_btwn_request=1,
                 initiator_name=None, folder_path='', base_url=None):
        self.output_path = os.path.join(folder_path, f'association_output_{datetime.datetime.now().strftime("%Y%m%d_%H%M%S")}.csv')
        super().__init__(associate_q=associate_q, stop_event=stop_event, owner_id=owner_id,
                         is_gcp=False, category_id=category_id, time_btwn_request=time_btwn_request,
                         initiator_name=initiator_name,
                         logger_config={'dir_name': os.path.basename(os.path.dirname(folder_path)),
                                        'folder_name': os.path.basename(folder_path)},
                         base_url=base_url)

    def handle_results(self, message, asset_id, pixel_dict, bad_association):
        bad_association = super().handle_results(message, asset_id, pixel_dict, bad_association)
        element = {
                   'asset_code': [asset_id],
                   'wiliot_code': pixel_dict['pixel_id'],
                   'is_associated': [int(message['status_code'] // 100) == 2],
                   'associate_status_code': [message['status_code']],
                   'associate_msg': [message['data']],}
        df = pd.DataFrame(element)
        try:
            if os.path.isfile(self.output_path):
                df.to_csv(self.output_path, mode='a', header=False, index=False)
            else:
                df.to_csv(self.output_path, index=False)
        except Exception as e:
            self.logger.warning(f'AssociatorProcess: Could NOT save the following data due to {e}\n{df}')
        # re-run bad association only if server error:
        re_try_association = []
        for b in bad_association:
            if int(message['status_code'] / 100) in [5, 9]:
                re_try_association.append(b)
        
        return re_try_association


class ContinuousScanning:
    def __init__(self, host='', port=23, user_configs=None, stop_event=None):
        self.user_configs = user_configs
        run_name = self.user_configs.get('run_name', 'test').replace('.csv', '')
        logger_path, self.logger = set_logger('CognexNetworkCont', folder_name=f'{run_name}_{datetime.datetime.now().strftime("%Y%m%d_%H%M%S")}', file_name=run_name)
        self.stop_event = stop_event
        self.output_path = logger_path.replace('.log', '.csv')
        self.logger.info(f'the output csv file will be generate on: {self.output_path}')
        self.codes = {'wiliot': [], 'asset': []}
        self.last_wiliot_asset_coupling = []
        self.summary = {'#Scanned': 0, '#Good': 0, '#Estimated': 0, '#BadPixels': 0, '#GoodPixels': 0,
                        '1stWiliotCode': '-', 'LastWiliotCode': '-', 'TimeFromLastGoodScanned': 0.0, 'last_good_timestamp': datetime.datetime.now().timestamp(),
                        '#ReconnectsAttempts': 0}
        try:
            scanner_kwargs = {'timeout': 0.1, 'log_name': self.logger.name, 'device_name': user_configs.get('device_name')}
            if 'csv_file' in user_configs:
                scanner_kwargs['sim_csv_path'] = user_configs['csv_file']
                scanner_kwargs['enable'] = False
            if host:
                scanner_kwargs['host'] = host
                scanner_kwargs['telnet_port'] = int(port)
                self.scanner = CognexNetwork(**scanner_kwargs)
            else:
                scanner_kwargs['port'] = port
                self.scanner = CognexDataMan(**scanner_kwargs)
        except Exception as e:
            self.logger.warning(f'could not connect to network: {host}::{port} scanner due to: {e}')
            raise e
        
        
        self.do_association = self.user_configs.get('do_association', False)
        self.association_handler, self.associate_q = None, None
        if self.do_association:
            try:
                queue_handler = QueueHandler()
                self.associate_q = queue_handler.get_multiprocess_queue(queue_max_size=500)
                initiator_name = os.environ.get('StationName')
                kwargs = {'associate_q': self.associate_q,
                          'stop_event': stop_event,
                          'owner_id': self.user_configs.get('owner_id',''),
                          'category_id': self.user_configs.get('category_id',''),
                          'initiator_name': initiator_name,
                          'folder_path': os.path.dirname(logger_path),
                          'base_url': self.user_configs.get('base_url')
                          }
                # check cloud connection:
                self.association_handler = multiprocessing.Process(target=AssociatorProcess,
                                                                    kwargs=kwargs)

            except Exception as e:
                self.logger.warning(f'exception during TagAssociation init: {e}')
                raise e
            
    def run(self):
        if self.association_handler is not None:
                self.association_handler.start()
        time_per_loc = self.user_configs.get('time_per_loc', 1)
        last_received = time.time()
        while not self.is_stopped():
            try:
                code_types = self.scan_per_location(time_per_loc=time_per_loc)
                if len(code_types['all_codes']) == 0:
                    if time.time() - last_received > MACHINE_PING_TIMEOUT:
                        self.logger.warning(f'No codes were scanned for more than {MACHINE_PING_TIMEOUT} seconds, trying to communicate with the scanner')
                        trigger_stat = self.scanner.get_device_property('TRIGGER')
                        self.logger.info(f'trigger status: {trigger_stat}')
                        last_received = time.time()
                
                else:
                    # scanned codes:
                    
                    last_received = time.time()
                    code_types = self.arrange_codes(code_types=code_types)
                    new_df = pd.DataFrame({'timestamp': [datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S")],
                                        'is_valid': [code_types['is_valid']],
                                        'wiliot_code': code_types['wiliot_code'], 
                                        'asset_code': code_types['asset_code'], 
                                            'all_codes': [code_types['all_codes']]})
                    self.save_file(new_df)

                    estimated_codes_out = self.estimate_missing_scanned_codes(code_types=code_types)
                    self.update_summary(code_types=code_types, estimated_codes=estimated_codes_out)
                    self.add_to_association(code_types_list=estimated_codes_out + [code_types])
                
                self.summary['TimeFromLastGoodScanned'] = datetime.datetime.now().timestamp() - self.summary['last_good_timestamp']
                self.logger.info(f'**********  SUMMARY: {self.summary} **********')
            except Exception as e:
                self.logger.warning(f'got exception during ContinuousScanning run: {e}')
                self.summary['#ReconnectsAttempts'] += 1
                st = self.scanner.reconnect()
                if not st:
                    time.sleep(1)

        self.stop()
    
    def scan_per_location(self, time_per_loc):
        code_types = {'wiliot_code': [], 'asset_code': [], 'all_codes': '',
                      'is_valid': False, 'timestamp': datetime.datetime.now().timestamp()}
        t_now = time.time()
        self.logger.info(f'Scanning for new location for {time_per_loc} seconds')
        while time.time() - t_now < time_per_loc:
            time.sleep(0.050)
            try:
                codes = self.scanner.read_response()
            except Exception as e:
                if 'timed out' not in str(e):
                    raise e
                codes = ''
            if codes == '' or codes == '\r\n':
                continue  # scan again
            self.logger.info(f'raw scanned ({len(codes.split("\r\n"))} msg): {codes}')
            codes = codes.replace('\r\n', '')
            # read something:
            code_types['timestamp'] = datetime.datetime.now().timestamp()
            code_types['all_codes'] = codes
            codes_list = codes.split(',')
            for c in codes_list:
                if is_wiliot_code(c):
                    code_types['wiliot_code'].append(c)
                elif self.user_configs.get('asset_str', '!!!') in c:
                    code_types['asset_code'].append(c)
            break
        
        return code_types

    def arrange_codes(self, code_types):
        code_types['is_valid'] = False
        # do clean up
        # wiliot_codes = []
        for c in code_types['wiliot_code']:
            if c not in self.codes['wiliot']:
                # wiliot_codes.append(c)
                self.codes['wiliot'].append(c)
            else:
                self.logger.warning(f'Wiliot code {c} was already detected')  # f'Wiliot code {c} was already detected, ignoring it'
                
        # asset_codes = []
        for c in code_types['asset_code']:
            if c not in self.codes['asset']:
                # asset_codes.append(c)
                self.codes['asset'].append(c)
            else:
                self.logger.warning(f'Asset code {c} was already detected')  # f'Asset code {c} was already detected, ignoring it'
        
        # code_types['wiliot_code'] = wiliot_codes
        # code_types['asset_code'] = asset_codes
        
        if len(code_types['wiliot_code']) == 1 and len(code_types['asset_code']) == 1:
            code_coupled = f"{code_types['wiliot_code'][0]} | {code_types['asset_code'][0]}"
            self.logger.info(f'Found New Wiliot-Asset combination: {code_coupled}')
            code_types['is_valid'] = True
        else:
            if len(code_types['wiliot_code']) > 1:
                self.logger.warning(f'More then 1 Pixels were detected: {code_types["wiliot_code"]}')
                code_types['wiliot_code'] = ['']
            if len(code_types['asset_code']) > 1:
                self.logger.warning(f'More then 1 Assets were detected: {code_types["asset_code"]}')
                code_types['asset_code'] = ['']
            if len(code_types['wiliot_code']) == 0:
                self.logger.warning(f'No Pixel was detected')
                code_types['wiliot_code'] = ['']
            if len(code_types['asset_code']) == 0:
                self.logger.warning(f'No Assets was detected')
                code_types['asset_code'] = ['']
        return code_types

    def update_summary(self, code_types, estimated_codes):
        self.summary['#Scanned'] += 1
        self.summary['#Good'] += int(code_types['is_valid'])
        try:
            if code_types['wiliot_code'][0] != '':
                if self.summary['1stWiliotCode'] == '-':
                    self.summary['1stWiliotCode'] = code_types['wiliot_code'][0]
                    self.summary['LastWiliotCode'] = code_types['wiliot_code'][0]
                elif code_types['wiliot_code'][0].split('T')[0] == self.summary['1stWiliotCode'].split(',')[-1].split('T')[0]:
                    self.summary['LastWiliotCode'] = ','.join(self.summary['LastWiliotCode'].split(',')[:-1] + [code_types['wiliot_code'][0]])
                else:  # from the different reel
                    self.summary['1stWiliotCode'] = self.summary['1stWiliotCode'] + ',' + code_types['wiliot_code'][0]
                    self.summary['LastWiliotCode'] = self.summary['LastWiliotCode'] + ',' + code_types['wiliot_code'][0]
            if self.summary['1stWiliotCode'] != '-':
                self.summary['#GoodPixels'] = sum([int(y.split('T')[1]) - int(x.split('T')[1]) + 1 for x, y in zip(self.summary['1stWiliotCode'].split(','), self.summary['LastWiliotCode'].split(','))])
        except Exception as e:
            self.logger.warning(f'Could not update summary for first and last Wiliot codes {code_types["wiliot_code"]} due to: {e}')
            
        if code_types['is_valid']:
            self.summary['last_good_timestamp'] = code_types['timestamp']
        self.summary['#Estimated'] += len(estimated_codes)

    def save_file(self, df):
        try:
            if os.path.isfile(self.output_path):
                df.to_csv(self.output_path, mode='a', header=False, index=False)
            else:
                df.to_csv(self.output_path, index=False)
        except Exception as e:
            self.logger.warning(f'Could NOT save the following data due to {e}\n{df}')
        return
    
    def estimate_missing_scanned_codes(self, code_types):
        estimated_codes_out = []
        pop_ind_list = [0]
        try:
            self.last_wiliot_asset_coupling.append(code_types)
            while len(self.last_wiliot_asset_coupling) > 4:
                self.last_wiliot_asset_coupling.pop(0)
            if len(self.last_wiliot_asset_coupling) == 4:
                ind_invalid = [i for i, x in enumerate(self.last_wiliot_asset_coupling) if not x['is_valid']]
                if 0 < len(ind_invalid) < 3:
                    # try to estimate codes:
                    if ind_invalid == [1] or ind_invalid == [2]:
                        i = ind_invalid[0]
                        pre_tag_num = int(self.last_wiliot_asset_coupling[i-1]['wiliot_code'][0].split('T')[1])
                        post_tag_num = int(self.last_wiliot_asset_coupling[i+1]['wiliot_code'][0].split('T')[1])
                        same_reel_id = self.last_wiliot_asset_coupling[i-1]['wiliot_code'][0].split('T')[0] == self.last_wiliot_asset_coupling[i+1]['wiliot_code'][0].split('T')[0]
                        if self.last_wiliot_asset_coupling[i]['asset_code'][0] != '' and same_reel_id:
                            if abs(post_tag_num - pre_tag_num) == 2:
                                # estimate the missing Wiliot code
                                pre_wiliot_code = self.last_wiliot_asset_coupling[i-1]['wiliot_code'][0]
                                offset_diff = 1 if post_tag_num > pre_tag_num else -1
                                estimated_wiliot_code = pre_wiliot_code.split('T')[0] + 'T' + str(pre_tag_num + offset_diff).zfill(4)
                                self.last_wiliot_asset_coupling[i]['wiliot_code'][0] = estimated_wiliot_code
                                self.last_wiliot_asset_coupling[i]['is_valid'] = True
                                self.logger.info(f'Estimating missing Wiliot-Asset code: {estimated_wiliot_code} | {self.last_wiliot_asset_coupling[i]["asset_code"][0]}')
                                estimated_codes_out.append(self.last_wiliot_asset_coupling[i])
                            elif abs(post_tag_num - pre_tag_num) == 1:
                                self.summary['#BadPixels'] += 1
                                pop_ind_list = [i]  # remove bad pixels ind

                    elif ind_invalid == [1, 2]:
                        pre_tag_num = int(self.last_wiliot_asset_coupling[0]['wiliot_code'][0].split('T')[1])
                        post_tag_num = int(self.last_wiliot_asset_coupling[3]['wiliot_code'][0].split('T')[1])
                        same_reel_id = self.last_wiliot_asset_coupling[0]['wiliot_code'][0].split('T')[0] == self.last_wiliot_asset_coupling[3]['wiliot_code'][0].split('T')[0]
                        if self.last_wiliot_asset_coupling[1]['asset_code'][0] != '' and self.last_wiliot_asset_coupling[2]['asset_code'][0] != '' and same_reel_id:
                            if abs(post_tag_num - pre_tag_num) == 3:
                                # estimate the missing Wiliot code
                                pre_wiliot_code = self.last_wiliot_asset_coupling[0]['wiliot_code'][0]
                                signed_diff = 1 if post_tag_num > pre_tag_num else -1
                                for i in [1, 2]:
                                    estimated_wiliot_code = pre_wiliot_code.split('T')[0] + 'T' + str(pre_tag_num + (i*signed_diff)).zfill(4)
                                    self.last_wiliot_asset_coupling[i]['wiliot_code'][0] = estimated_wiliot_code
                                    self.last_wiliot_asset_coupling[i]['is_valid'] = True
                                    self.logger.info(f'Estimating missing Wiliot-Asset code: {estimated_wiliot_code} | {self.last_wiliot_asset_coupling[i]["asset_code"][0]}')
                                    estimated_codes_out.append(self.last_wiliot_asset_coupling[i])
                            elif abs(post_tag_num - pre_tag_num) == 1:
                                self.summary['#BadPixels'] += 2
                                pop_ind_list = [1,2]  # remove bad pixels ind

                for i, i_p in enumerate(pop_ind_list):
                    self.last_wiliot_asset_coupling.pop(i_p-i)  # remove the first element or bad pixel elements
        except Exception as e:
            self.logger.warning(f'Could not estimate missing scanned codes for {self.last_wiliot_asset_coupling} due to: {e}')

        return estimated_codes_out

    def add_to_association(self, code_types_list):
        if not self.do_association:
            return
        
        for c in code_types_list:
            if c['is_valid']:
                if self.associate_q.full():
                    self.logger.warning(f'Could not add to association queue since queue is full discard: {c}')
                    return
                c['asset_code'] = [f"{c['asset_code'][0]}-{round(c['timestamp'])}"]
                self.associate_q.put(c, block=False)
    
    def is_stopped(self):
        if self.stop_event is not None:
            return self.stop_event.is_set()
        return False
    
    def stop(self):
        self.scanner.trigger_off()
        time.sleep(0.1)
        self.scanner.close_port()
        if self.association_handler is not None:
                self.association_handler.join(10)
                if self.association_handler.is_alive():
                    self.logger.warning('associator process is still running')
        self.logger.info(f'ContinuousScanning run was completed, please check: {self.output_path}')


if  __name__ == '__main__':
    import argparse
    import keyboard
    import threading

    stop_event = multiprocessing.Event()
    def exit_func():
        while True:  # making a loop
            time.sleep(0.5)
            try:
                if keyboard.is_pressed('q'):
                    print('Exit app!')
                    stop_event.set()
                    break  # finishing the loop

            except Exception as e:
                print(e)
                break

    parser = argparse.ArgumentParser(description='Run Continuous Scanner')
    parser.add_argument('-hn', '--host', default=None, help='host ip address')
    parser.add_argument('-pn', '--port', default=None, help='port number')
    parser.add_argument('-r', '--run_name', default=None, help='the run name for logging the output csv file')
    parser.add_argument('-t', '--time_per_loc', default=None, help='time per each location/box/frame')
    parser.add_argument('-a', '--asset_str', default=None, help='unique string inside the asset code')
    parser.add_argument('-d', '--do_association', default=0, help='True if association is needed')
    parser.add_argument('-o', '--owner_id', default=None, help='The owner id number')
    parser.add_argument('-c', '--category_id', default=None, help='The owner id number')
    parser.add_argument('-b', '--base_url', default=None, help='The base url for instance https://api.us-east-2.prod.wiliot.cloud')
    parser.add_argument('-dn', '--device_name', default=None, help='The name of the scanner device, e.g. DM-1234')

    args = parser.parse_args()
    user_configs={'run_name': args.run_name, 
                  'time_per_loc': float(args.time_per_loc), 
                  'asset_str': args.asset_str,
                  'do_association': args.do_association,
                  'owner_id': args.owner_id,
                  'category_id': args.category_id,
                  'base_url': args.base_url,
                  'device_name': args.device_name
                  }
    cs =ContinuousScanning(host=args.host, port=args.port, 
                           stop_event=stop_event, user_configs=user_configs)
    exit_thread = threading.Thread(target=exit_func, args=())
    exit_thread.start()
    cs.run()
    print('done')


from dash import html, dcc
import dash_bootstrap_components as dbc

from wiliot_core.utils.utils import IS_PRIVATE_INSTALLATION
from wiliot_testers.failure_analysis_tester.configs_gui import RELAY_CONFIG

ELEMENT_WIDTH = "300px"
HALF_ELEMENT_WIDTH = "150px"
BUTTON_WIDTH = 30
CELL_STYLE = {"width": ELEMENT_WIDTH, "display": "flex", "justifyContent": "center",
              "alignItems": "center", "flexDirection": "column", "gap": "10px"}
BUTTON_STYLE = {"width": "200px", "height": "20px", "fontSize": "14px",
                "display": "flex", "alignItems": "center", "justifyContent": "center"}
TEST_OPTIONS = []
TEST_OPTIONS.append({'label': 'IV Curve', 'value': 'IV Curve'})
if IS_PRIVATE_INSTALLATION:
    TEST_OPTIONS.append({'label': 'Capacitance', 'value': 'Capacitance'},)
    # TEST_OPTIONS.append({'label': 'Voltage Drop', 'value': 'Voltage Drop'},)
    # TEST_OPTIONS.append({'label': 'External Power Source', 'value': 'External Power Source'})

TAG_OPTIONS = [
    {'label': 'E4', 'value': 'E4'},
    {'label': 'E2', 'value': 'E2'}] 


def get_tag_alias(disabled=False):
    return html.Div(
        [html.Label("Tag Alias"),
         dbc.Input(id="tag_alias", type="text", disabled=disabled)],
        style={"width": ELEMENT_WIDTH})


def get_folder_name(disabled=False):
    return html.Div(
        [html.Label("Folder Run Name"),
         dbc.Input(id="folder_name", type="text", disabled=disabled)],
        style={"width": ELEMENT_WIDTH})


def get_comment(disabled=False):
    return html.Div(
        [html.Label("Comment (optional)"),
         dbc.Input(id="test_comment", type="text", disabled=disabled)],
        style={"width": ELEMENT_WIDTH})


def get_test_and_tag_type(disabled=False):
    return html.Div([
        html.Div([
        html.Label("Test Type"),
        dcc.Dropdown(
            disabled=disabled,
            id='test_type',
            options=TEST_OPTIONS,
            value='IV Curve'),],
        style={"width": HALF_ELEMENT_WIDTH}), 
        html.Div([
        html.Label("Tag Type"),
        dcc.Dropdown(
            disabled=disabled,
            id='tag_type',
            options=TAG_OPTIONS,
            value='E4'),],
        style={"width": HALF_ELEMENT_WIDTH})
        ], style={'display': 'flex', 'flexDirection': 'row', 'gap': '10px'})


def get_run_finish_buttons(disabled=False):
    return html.Div([
        dbc.Button("Run Test", id="run_test", color="success",
                   n_clicks=0, style={"width": "200px"}, disabled=disabled),
        dbc.Button("Finish Test", id="finish_test",
                   color="danger", n_clicks=0, style={"width": "200px"}, disabled=disabled),
    ], style=CELL_STYLE)


def get_test_checkbox():
    options = [{'label': key, 'value': key, 'disabled': False} for key in RELAY_CONFIG.keys()]
    return html.Div([
        dcc.Checklist(options=options, value=list(RELAY_CONFIG.keys()), id="test_fields",),
    ], style={"width": ELEMENT_WIDTH, "gap": "10px", "display": "flex", "alignItems": "center"})


def get_test_result(disabled=False):
    return html.Div([
        html.Label("Results"),
        dbc.Textarea(id="test_result", rows=4, disabled=True)
    ], style={"width": ELEMENT_WIDTH})


def get_smu_config(default_value=''):
    return html.Div([
        dbc.Input(id="smu_address", type="text", value=default_value),
        dbc.Button("Set SMU address", id="smu_address_button",
                   color="info", n_clicks=0, style=BUTTON_STYLE),
        dbc.Button("Open Output Folder", id="open_output_folder",
            color="info", className="mb-2", style=BUTTON_STYLE, disabled=True),

    ], style={"width": ELEMENT_WIDTH, "display": "flex", "justifyContent": "center",
              "alignItems": "center", "flexDirection": "column", "gap": "10px"})



def get_plot_selection(disabled=False):
    return html.Div([
        html.Label("Plot Selection"),
        dcc.Dropdown(
            id='plot-selection',
            options=['Empty'] + list(RELAY_CONFIG.keys()),
            value='Empty',
            disabled=disabled),
    ], style={"width": ELEMENT_WIDTH})




"""
Copyright (c) 2016- 2025, Wiliot Ltd. All rights reserved.

Redistribution and use of the Software in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

  2. Redistributions in binary form, except as used in conjunction with
  Wiliot's Pixel in a product or a Software update for such product, must reproduce
  the above copyright notice, this list of conditions and the following disclaimer in
  the documentation and/or other materials provided with the distribution.

  3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
  may be used to endorse or promote products or services derived from this Software,
  without specific prior written permission.

  4. This Software, with or without modification, must only be used in conjunction
  with Wiliot's Pixel or with Wiliot's cloud service.

  5. If any Software is provided in binary form under this license, you must not
  do any of the following:
  (a) modify, adapt, translate, or create a derivative work of the Software; or
  (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
  discover the source code or non-literal aspects (such as the underlying structure,
  sequence, organization, ideas, or algorithms) of the Software.

  6. If you create a derivative work and/or improvement of any Software, you hereby
  irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
  royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
  right and license to reproduce, use, make, have made, import, distribute, sell,
  offer for sale, create derivative works of, modify, translate, publicly perform
  and display, and otherwise commercially exploit such derivative works and improvements
  (as applicable) in conjunction with Wiliot's products and services.

  7. You represent and warrant that you are not a resident of (and will not use the
  Software in) a country that the U.S. government has embargoed for use of the Software,
  nor are you named on the U.S. Treasury Department’s list of Specially Designated
  Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
  You must not transfer, export, re-export, import, re-import or divert the Software
  in violation of any export or re-export control laws and regulations (such as the
  United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
  and use restrictions, all as then in effect

THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
(SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
(A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
(B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
(C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""

import sys
import time
import logging
import pandas as pd
from dash import Dash, html, dcc, dash_table, ctx
import dash_bootstrap_components as dbc
from dash.dependencies import Input, Output, State
import threading
import traceback
from werkzeug.serving import make_server

from wiliot_core.utils.utils import GetApiKey
from wiliot_testers.association_tester.modules.gui_module_components import get_rescan_modal, get_upload_modal
from wiliot_testers.utils.upload_to_cloud_api import upload_to_cloud_api,TesterType


HOST = '127.0.0.1'
PORT = 8070
COLUMNS = ['location', 'wiliot_code', 'asset_code', 'timestamp', 'scan_status',
           'is_associated', 'associate_status_code', 'adv_address', 'n_packets', 'external_id', 'is_success']


class AssociationAndVerificationGUI(object):
    def __init__(self, logger_name='root', get_data_func=None, get_stat_func=None, stop_event=None, rescan_event=None, gui_done_event=None, tag_association=None, app_config=None):
        self.logger = logging.getLogger(logger_name)
        self.app_config = app_config if app_config is not None else {}
        self.stop_button_status = False
        self.is_running = True
        self.common_run_name = self.app_config.get('common_run_name', '')
        self.stop_event = stop_event
        self.rescan_event = rescan_event
        self.gui_done_event = gui_done_event
        self.rescan_modal_text = ''
        self.was_upload_modal_open = False
        self.tag_association = tag_association
        self.demo_index = 0
        self.is_done_index = 0
        self.get_data = get_data_func if get_data_func is not None else self.get_data_demo
        self.get_stat = get_stat_func if get_stat_func is not None else self.calc_stat_demo
        self.results = pd.DataFrame()
        self.dash_app = Dash(__name__, external_stylesheets=[dbc.themes.BOOTSTRAP])
        self.dash_app.title = 'Wiliot Association & Verification Tester'

        self.all_cards = {'location_number': 'secondary',
                          'scan_success': 'primary',
                          'association_success': 'success',
                          'responding_rate': 'info'}

        self.generate_inlay()

        self.server = self.dash_app.server
        self.init_callbacks()

    def generate_inlay(self):
        def header(name):
            title = html.H2(name, style={"margin-top": 5})
            logo = html.Img(
                src='https://www.wiliot.com/src/uploads/Wiliotlogo.png', style={"float": "right", "height": 50}
            )
            return dbc.Row([dbc.Col(title, md=9), dbc.Col(logo, md=3)])

        def card_content(card_id):
            return [html.Div(id=card_id),
                    html.P(card_id.capitalize().replace('_', ' '), className="card-text")]

        # Card components
        cards = [dbc.Card(card_content(card_name), body=True, color=card_color, inverse=True)
                 for card_name, card_color in self.all_cards.items()]

        # table
        data_table = dash_table.DataTable(
            id='results_table',
            columns=[{'name': col, 'id': col} for col in COLUMNS],
            data=[],
            filter_action="native",
            sort_action="native",
            sort_mode="multi",
            style_cell={'textAlign': 'left', 'padding': '5px'},
            style_header={'fontWeight': 'bold'},
        )

        # buttons
        stop_button = dbc.Button("Stop", id="stop-btn", className="mb-3", color="danger", n_clicks=0)
        pause_button = dbc.Button("Pause", id="pause-btn", className="mb-3", color="warning", n_clicks=0)
        is_r2r_printer = "printer" in self.tag_association.user_inputs["r2r_type"]

        self.dash_app.layout = dbc.Container(
            [
                html.Br(),
                header("Wiliot Association & Verification Tester"),
                html.Hr(),
                dbc.Row([dbc.Col(card) for card in cards]),
                html.Br(),
                get_rescan_modal(is_r2r_printer),
                get_upload_modal(),
                html.Div([html.Div(id='app-title', children=f'{self.common_run_name}: 0 success out of 0',
                                   style={'fontSize': 20, 'fontWeight': 'bold'})]),
                html.Br(),
                html.Div([stop_button, pause_button, html.Div(id='app-state', children='Application is running')]),
                html.Br(),
                html.Div([html.H3("Results"), data_table, html.Div(id='results_table_container')]),
                dcc.Interval(
                    id='interval-component',
                    interval=1 * 1000,  # in milliseconds
                    n_intervals=0),
                html.Footer('Copyright (c) 2024 Wiliot')
            ],
            fluid=False,
        )

    def init_callbacks(self):
        @self.dash_app.callback(
            Output("scan-results", "value", allow_duplicate=True),
            Input("reprint-btn", "n_clicks"),
            prevent_initial_call=True
        )
        def reprint(n_clicks):
            self.tag_association.reprint_previous_label()
            return "Reprinting last label"

        @self.dash_app.callback(
            Output("scan-results", "value", allow_duplicate=True),
            Input("rescan-btn", "n_clicks"),
            prevent_initial_call=True
        )
        def rescan(n_clicks):
            try:
                self.tag_association.scan()
                self.tag_association.last_exception_for_print = ''
            except ValueError as e:
                self.tag_association.last_exception_for_print = str(e)
            self.rescan_modal_text = self.tag_association.get_scan_str()
            return self.rescan_modal_text

        @self.dash_app.callback(
            Output('rescan-modal', 'is_open', allow_duplicate=True),
            Input("continue-rescan-modal", "n_clicks"),
            prevent_initial_call=True
        )
        def rescan_continue(n_clicks):
            self.rescan_event.clear() 
            return False

        @self.dash_app.callback(
            Output('rescan-modal', 'is_open', allow_duplicate=True),
            Input("stop-rescan-modal", "n_clicks"),
            prevent_initial_call=True
        )
        def rescan_stop(n_clicks):
            self.stop_event.set() 
            self.rescan_event.clear() 
            return False

        @self.dash_app.callback(
            Output("connect-btn", "children"),
            Output("rescan-btn", "disabled",),
            Output("continue-rescan-modal", "disabled",),
            Input("connect-btn", "n_clicks"),
            prevent_initial_call=True
        )
        def rescan_disconnect(n_clicks):
            success, act_str = self.tag_association.disconnect_or_reconnect_scanner()
            if success:
                btn_str = 'Reconnect Scanner' if act_str[0]=='D' else 'Disconnect Scanner'
                self.rescan_modal_text = f'Scanner {act_str} Succeeded'
                rescan_button_disable = act_str[0]=='D'
            else:
                btn_str = act_str[:-3] + ' Scanner'
                self.rescan_modal_text = f'Scanner {act_str} Failed'
                rescan_button_disable = act_str[0]=='R'
            continue_button_disable = rescan_button_disable
            return btn_str, rescan_button_disable, continue_button_disable
        
        @self.dash_app.callback(
            Output('upload-modal', 'is_open', allow_duplicate=True),
            Input("dismiss-btn-upload-modal", "n_clicks"),
            prevent_initial_call=True
        )
        def dismiss_upload(n_clicks):
            self.gui_done_event.set()
            return False
        
        @self.dash_app.callback(
            Output('upload-modal', 'is_open', allow_duplicate=True),
            Output('results-upload-modal', 'value'),
            Input("upload-btn-upload-modal", "n_clicks"),
            State("env-dropdown-upload-modal", "value"),
            State("owner-id-upload-modal", "value"),
            prevent_initial_call=True
        )
        def upload_upload(n_clicks, env, owner_id):
            if owner_id is None or owner_id=='':
                return True, 'Please provide owner id'
            
            auth_gui_is_needed, _, _, _ = GetApiKey.check_file(owner_id, env)
            if auth_gui_is_needed:
                return True, 'Could not find api key for the provided owner id and env. Please check your input, or add api key to the config file'
            err_str = ''
            try:
                self.upload_data_to_cloud(env, owner_id)
                self.gui_done_event.set()
                return False, ''
            except Exception as e:
                err_str = f"data failed to upload for {self.common_run_name} due to: {e}, traceback :{traceback.format_exc()}"
                self.logger.warning(err_str)

            return True, err_str

        @self.dash_app.callback(
            [Output(card_name, 'children') for card_name in self.all_cards.keys()] + [Output('results_table', 'data'),
                                                                                      Output('app-title', 'children'),
                                                                                      Output('rescan-modal', 'is_open', allow_duplicate=True),
                                                                                      Output('upload-modal', 'is_open', allow_duplicate=True),
                                                                                      Output("scan-results", "value", allow_duplicate=True),],
            [Input('interval-component', 'n_intervals'), State('rescan-modal', 'is_open'), State('upload-modal', 'is_open')],
            prevent_initial_call=True
        )
        def update_data(n_intervals, rescan_modal_visibility, upload_modal_visibility):
            # if self.gui_done_event.is_set():
            #     if self.is_done_index == 3:
            #         self.logger.info('Wiliot A&V has been stopped')
            #         sys.exit(-1)
            #     self.is_done_index += 1
            self.results = self.get_data()
            new_stat = self.get_stat()

            n_locations = html.H2(new_stat['n_locations'], className="card-title")
            scan_success = html.H2(new_stat['scan_success'], className="card-title")
            association_success = html.H2(new_stat['association_success'], className="card-title")
            responding_rate = html.H2(new_stat['responding_rate'], className="card-title")

            msg = f'{self.common_run_name}: {new_stat["n_success"]} success out of {new_stat["n_locations"]}'
            style = {}
            if self.stop_event is not None and self.stop_event.is_set():
                msg += ' \n Application Was Stopped'
                style = {"fontSize": "40px", "backgroundColor": "red", 'whiteSpace': 'pre-line'}
                if not self.was_upload_modal_open:
                    upload_modal_visibility = True
                    self.was_upload_modal_open = True

            if self.rescan_event is not None and self.rescan_event.is_set() and not rescan_modal_visibility:
                self.rescan_modal_text = self.tag_association.get_scan_str()
                rescan_modal_visibility = True

            return [n_locations, scan_success, association_success, responding_rate, self.results.to_dict("records"),
                    html.Div(msg,style=style), rescan_modal_visibility, upload_modal_visibility, self.rescan_modal_text]

        @self.dash_app.callback(
            [Output('app-state', 'children'), Output('pause-btn', 'children'), Output('pause-btn', 'color')],
            [Input('stop-btn', 'n_clicks'), Input('pause-btn', 'n_clicks')]
        )
        def buttons(n_clicks1, n_click2):
            msg = 'Application is running'
            pause_str = 'Pause'
            pause_style = 'warning'
            style = {}

            if "stop-btn" == ctx.triggered_id:
                self.stop_button_status = True
                self.is_running = False
                msg = 'Application was stopped'
                style = {"fontSize": "40px", "backgroundColor": "red"}
            elif "pause-btn" == ctx.triggered_id:
                self.is_running = (not self.is_running)
                if not self.is_running:
                    msg = 'Application was paused'
                    pause_str = 'Run'
                    pause_style = 'success'

            return [html.Div(msg, style=style), pause_str, pause_style]

    def run_app(self):
        self.server = make_server(HOST, PORT, self.dash_app.server)
        logging.getLogger('werkzeug').setLevel(logging.ERROR)
        self.server.serve_forever()

    def is_stopped_by_user(self):
        return self.stop_button_status

    def is_app_running(self):
        return self.is_running

    @staticmethod
    def get_url():
        return f"http://{HOST}:{PORT}/"

    def calc_stat_demo(self):
        df_in = self.results
        stat_out = {}
        rel_data = df_in.loc[~df_in['location'].isna()]
        n_location = len(rel_data)
        stat_out['n_locations'] = str(n_location)
        stat_out['n_tags_outside_test'] = str(len(df_in) - n_location)
        if n_location:
            stat_out['scan_success'] = f'{round(rel_data["scan_status"].sum() / n_location * 100, 2)}%'
            stat_out['association_success'] = f'{round(rel_data["is_associated"].sum() / n_location * 100, 2)}%'
            stat_out['responding_rate'] = f'{round(rel_data["n_packets"].notna().sum() / n_location * 100, 2)}%'

        return stat_out

    def get_data_demo(self):
        self.demo_index += 1
        path = "C:/Users/shunit/Downloads/association_and_verification_log_20240414_165814_results_df.csv"
        df = pd.read_csv(path)
        return df.iloc[0:self.demo_index]
    
    def upload_data_to_cloud(self, env, owner_id):
        # make sure user have the api key:
        run_data_path = self.app_config.get('run_data_path', f'{self.common_run_name}@run_data.csv')
        packets_data_path = self.app_config.get('packets_data_path', f'{self.common_run_name}@packets_data.csv')
        upload_to_cloud_api(batch_name=self.common_run_name, 
                            tester_type=TesterType.ASSOCIATION_AND_VERIFICATION_TEST,
                            run_data_csv_name=run_data_path, 
                            packets_data_csv_name=packets_data_path,
                            is_path=True,
                            env=env,
                            owner_id=owner_id)
        self.logger.info(f"data uploaded to cloud: {self.common_run_name}")


if __name__ == '__main__':
    import multiprocessing
    import webbrowser
    from wiliot_testers.association_tester.modules.association_module import TagAssociation
    tag_association = TagAssociation.__new__(TagAssociation)
    tag_association.user_inputs = {'r2r_type': 'zebra_printer'}

    rescan_event = multiprocessing.Event()
    t = AssociationAndVerificationGUI(tag_association=tag_association, rescan_event=rescan_event, app_config={'common_run_name': 'test_run'})
    plot_thread = threading.Thread(target=t.run_app, daemon=True, args=())
    plot_thread.start()
    webbrowser.open(t.get_url())
    while True:
        if not plot_thread.is_alive():
            break
        time.sleep(10)



"""
Copyright (c) 2016- 2025, Wiliot Ltd. All rights reserved.

Redistribution and use of the Software in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

  2. Redistributions in binary form, except as used in conjunction with
  Wiliot's Pixel in a product or a Software update for such product, must reproduce
  the above copyright notice, this list of conditions and the following disclaimer in
  the documentation and/or other materials provided with the distribution.

  3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
  may be used to endorse or promote products or services derived from this Software,
  without specific prior written permission.

  4. This Software, with or without modification, must only be used in conjunction
  with Wiliot's Pixel or with Wiliot's cloud service.

  5. If any Software is provided in binary form under this license, you must not
  do any of the following:
  (a) modify, adapt, translate, or create a derivative work of the Software; or
  (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
  discover the source code or non-literal aspects (such as the underlying structure,
  sequence, organization, ideas, or algorithms) of the Software.

  6. If you create a derivative work and/or improvement of any Software, you hereby
  irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
  royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
  right and license to reproduce, use, make, have made, import, distribute, sell,
  offer for sale, create derivative works of, modify, translate, publicly perform
  and display, and otherwise commercially exploit such derivative works and improvements
  (as applicable) in conjunction with Wiliot's products and services.

  7. You represent and warrant that you are not a resident of (and will not use the
  Software in) a country that the U.S. government has embargoed for use of the Software,
  nor are you named on the U.S. Treasury Department’s list of Specially Designated
  Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
  You must not transfer, export, re-export, import, re-import or divert the Software
  in violation of any export or re-export control laws and regulations (such as the
  United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
  and use restrictions, all as then in effect

THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
(SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
(A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
(B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
(C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""

import logging
import os.path
import datetime
import csv
from wiliot_api import TesterType
from wiliot_core import WiliotDir
from wiliot_tools.utils.wiliot_gui.wiliot_gui import WiliotGui, popup_message

from wiliot_testers.utils.upload_to_cloud_api import upload_to_cloud_api
from wiliot_testers.wiliot_tester_tag_result import FailureCodes

MAX_FILE_SIZE = 15 * 10 ** 6  # 15 Mb


def create_summary_message(is_upload=True, data_folder_name=''):
    """
    summary log
    :param is_upload:
    :type is_upload: bool or None
    :return:
    :rtype:
    @param data_folder_name:
    @type data_folder_name:  str
    """
    sum_msg_prefix = '\n\n\n***********************************************\n'
    sum_msg_suffix = '***********************************************\n'
    sum_msg_upload = f'********* UPLOAD: files were uploaded from {data_folder_name} *********\n'
    sum_msg = sum_msg_prefix
    if is_upload is not None:
        sum_msg += sum_msg_upload
        if not is_upload:
            sum_msg = sum_msg.replace('were', 'were NOT')
    sum_msg += sum_msg_suffix
    return sum_msg


class UploadTestersData(object):
    def __init__(self):
        """
        This function check the selected user files and upload them to the cloud.
        If the file exceed the size limit it split it
        """
        self.values = None
        self.logger = None

        self.get_files_path()
        if self.values is None:
            print('user exited the program')
            return

        self.set_logging()
        file_status = self.check_files()
        if not file_status:
            raise Exception(f'invalid file inputs: {self.values["run_data_file"]}, {self.values["packets_data_file"]}')

        # check file size:
        self.all_run_names = [self.values['run_data_file']]
        self.all_packets_names = [self.values['packets_data_file']]
        self.tester_type = self.values['tester_type']
        file_size = os.stat(self.values['packets_data_file']).st_size
        if file_size > MAX_FILE_SIZE and 'offline' in self.tester_type.lower():
            self.split_large_file(run_data=self.values['run_data_file'], packet_data=self.values['packets_data_file'],
                                  size=file_size)

    def get_files_path(self):
        """
        opens GUI for selecting a file and returns it
        """

        layout_dict = {
            'run_data_file': {'text': 'Choose run data file that you want to upload:', 'value': '',
                              'widget_type': 'file_input'},
            'packets_data_file': {'text': 'Choose packets data file that you want to upload:', 'value': '',
                                  'widget_type': 'file_input'},
            'env': {'text': 'Environment', 'value': "prod", 'widget_type': 'combobox',
                    'options': ('prod', 'test', 'dev')},
            'owner_id': {'text': 'Owner id:', 'value': '852213717688', 'widget_type': 'entry'},
            'tester_type': {'text': 'Tester type:', 'value': "OFFLINE_TEST", 'widget_type': 'combobox',
                            'options': tuple(TesterType.__members__.keys())}}

        gui = WiliotGui(params_dict=layout_dict, title='Upload testers data to cloud')
        values_out = gui.run()

        if values_out['run_data_file'] != '' or values_out['packets_data_file'] != '':
            self.values = values_out
        else:
            self.values = None

    def set_logging(self):
        """
        Sets up logging for the object, creating a log file with the current date and time as part of the filename.
        Returns: None
        """
        #  get()dirname from run_data or packets_data
        files_dir = os.path.dirname(self.values.get('run_data_file', self.values.get('packets_data_file')))
        if not files_dir:
            files_dir = os.path.join(WiliotDir().get_tester_dir("upload_to_cloud"), "logs")
            os.makedirs(files_dir, exist_ok=True)
        log_path = os.path.join(files_dir,
                                datetime.datetime.now().strftime("%d-%m-%Y_%H-%M-%S") + 'upload_to_cloud.log')
        logger = logging.getLogger('Manual Upload')
        logger.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(asctime)s,%(msecs)03d %(name)s %(levelname)s %(message)s', '%H:%M:%S')
        file_handler = logging.FileHandler(log_path)
        file_handler.setFormatter(formatter)
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(formatter)
        stream_handler.setLevel(logging.DEBUG)
        logger.addHandler(file_handler)
        logger.addHandler(stream_handler)
        self.logger = logger

    def check_files(self):
        """
        Checks if the run_data_file and packets_data_file are valid CSV files.
        Returns True if both files are valid, False otherwise.
        """
        if not os.path.isfile(self.values['run_data_file']) or not self.values['run_data_file'].endswith('.csv'):
            self.logger.warning('UPLOAD: run_data file format is not csv, please insert a csv file')
            self.logger.info(create_summary_message(False))
            return False
        if not os.path.isfile(self.values['packets_data_file']) or not self.values['packets_data_file'].endswith(
                '.csv'):
            self.logger.warning('UPLOAD: packets_data file format is not csv, please insert a csv file')
            self.logger.info(create_summary_message(False))
            return False
        return True

    def split_large_file(self, run_data, packet_data, size):

        """
        Splits a large packet_data CSV file into smaller parts that are each under the maximum allowed size.
        Duplicates the corresponding run_data CSV file for each part. Returns a list of file paths for each split part.
        """

        # Define a helper function to generate new file paths.
        def get_new_path_name(original_path, k, total, file_str):
            new_folder_name = os.path.dirname(original_path) + f'_{k}_out_of_{n_splits}'
            new_file_name = os.path.basename(original_path).replace(f'@{file_str}_data.csv',
                                                                    f'_{k}_out_of_{total}@{file_str}_data.csv')
            if not os.path.isdir(new_folder_name):
                os.makedirs(new_folder_name)
            return os.path.join(new_folder_name, new_file_name)

        # Split the file into smaller parts.
        n_splits = -(-size // MAX_FILE_SIZE)
        part = 1
        all_packets_path = []
        all_run_path = []
        # read
        f_read = open(packet_data, 'r')
        reader = csv.DictReader(f_read, delimiter=',')
        col_names = reader.fieldnames
        # write first file
        new_packet_path = get_new_path_name(packet_data, part, n_splits, 'packets')
        all_packets_path.append(new_packet_path)
        f_write = open(new_packet_path, 'w', newline='')
        writer = csv.DictWriter(f_write, fieldnames=col_names)
        writer.writeheader()
        last_loc = 0
        tested_list = []
        start_loc = -1
        missing_label_count = 0
        for i, row in enumerate(reader):
            row['common_run_name'] = os.path.basename(new_packet_path).split('@')[0]
            if os.stat(new_packet_path).st_size < MAX_FILE_SIZE or int(row['tag_run_location']) == last_loc:
                writer.writerow(row)
                last_loc = int(row['tag_run_location'])
                missing_label_count += int(row['fail_bin']) in (FailureCodes.MISSING_LABEL.value,
                                                               FailureCodes.END_OF_TEST.value)
            else:  # we need to split the file
                tested_list.append(last_loc - start_loc - missing_label_count)
                f_write.close()
                part += 1
                new_packet_path = get_new_path_name(packet_data, part, n_splits, 'packets')
                all_packets_path.append(new_packet_path)
                f_write = open(new_packet_path, 'w', newline='')
                writer = csv.DictWriter(f_write, fieldnames=col_names)
                writer.writeheader()
                row['common_run_name'] = os.path.basename(new_packet_path).split('@')[0]
                writer.writerow(row)
                start_loc = last_loc
                last_loc = int(row['tag_run_location'])
                missing_label_count = 0
        tested_list.append(last_loc - start_loc - missing_label_count)
        f_write.close()
        if part > n_splits:
            self.logger.info(
                f'Since the file cannot be split in the middle of location data, a mismatch happened and '
                f'the number of parts is bigger than the number of split (i.e you will have a file with '
                f'the suufix {part}_out_of{n_splits}')
        # Duplicate the run data for each part.
        for i in range(part):
            new_run_path = get_new_path_name(run_data, i + 1, n_splits, 'run')
            all_run_path.append(new_run_path)
            with open(run_data, 'r') as f_run_read:
                run_reader = csv.DictReader(f_run_read, delimiter=',')
                run_col_names = run_reader.fieldnames
                run_row = run_reader.__next__()
            run_row['common_run_name'] = os.path.basename(new_run_path).split('@')[0]
            run_row['total_run_tested'] = str(tested_list[i])
            with open(new_run_path, 'w', newline='') as f_run_write:
                run_writer = csv.DictWriter(f_run_write, fieldnames=run_col_names)
                run_writer.writeheader()
                run_writer.writerow(run_row)

        self.all_run_names = all_run_path
        self.all_packets_names = all_packets_path

    def upload_to_cloud(self):
        """
        Uploads files to a cloud API
        Returns True if all uploads were successful
        """
        all_status = []
        if self.tester_type.upper() not in TesterType.__members__.keys():
            raise Exception('unsupported tester type, please select type from the drop-down-list')
        for run_path, packet_path in zip(self.all_run_names, self.all_packets_names):
            try:
                upload_success = upload_to_cloud_api(batch_name=os.path.dirname(run_path),
                                                     tester_type=TesterType[self.tester_type],
                                                     run_data_csv_name=os.path.basename(run_path),
                                                     packets_data_csv_name=os.path.basename(packet_path),
                                                     env=self.values['env'],
                                                     is_batch_name_inside_logs_folder=False,
                                                     logger_=self.logger.name, owner_id=self.values['owner_id'])
            except Exception as e:
                self.logger.warning(f'UPLOAD: during upload_to_cloud_api an error occurred: {e}')
                upload_success = False

            if not upload_success:
                msg = "Run upload failed. Check exception error at the console and check Internet connection is " \
                      "available and upload logs manually"
                bg = 'red'
                popup_message(msg=msg, bg=bg)

            message = create_summary_message(is_upload=upload_success, data_folder_name=os.path.dirname(run_path))
            self.logger.info(message)
            all_status.append(upload_success)

        return all(all_status)


if __name__ == '__main__':
    # upload to cloud
    upload_obj = UploadTestersData()
    status = upload_obj.upload_to_cloud()

    print(f"done with status {status}")



"""
Copyright (c) 2016- 2025, Wiliot Ltd. All rights reserved.

Redistribution and use of the Software in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

  2. Redistributions in binary form, except as used in conjunction with
  Wiliot's Pixel in a product or a Software update for such product, must reproduce
  the above copyright notice, this list of conditions and the following disclaimer in
  the documentation and/or other materials provided with the distribution.

  3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
  may be used to endorse or promote products or services derived from this Software,
  without specific prior written permission.

  4. This Software, with or without modification, must only be used in conjunction
  with Wiliot's Pixel or with Wiliot's cloud service.

  5. If any Software is provided in binary form under this license, you must not
  do any of the following:
  (a) modify, adapt, translate, or create a derivative work of the Software; or
  (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
  discover the source code or non-literal aspects (such as the underlying structure,
  sequence, organization, ideas, or algorithms) of the Software.

  6. If you create a derivative work and/or improvement of any Software, you hereby
  irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
  royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
  right and license to reproduce, use, make, have made, import, distribute, sell,
  offer for sale, create derivative works of, modify, translate, publicly perform
  and display, and otherwise commercially exploit such derivative works and improvements
  (as applicable) in conjunction with Wiliot's products and services.

  7. You represent and warrant that you are not a resident of (and will not use the
  Software in) a country that the U.S. government has embargoed for use of the Software,
  nor are you named on the U.S. Treasury Department’s list of Specially Designated
  Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
  You must not transfer, export, re-export, import, re-import or divert the Software
  in violation of any export or re-export control laws and regulations (such as the
  United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
  and use restrictions, all as then in effect

THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
(SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
(A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
(B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
(C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""
from collections import Counter
import datetime
import os
import re
import sys
import time
import json
from math import nan
from queue import Queue
import matplotlib
import pandas as pd
import threading
import numpy as np
import webbrowser

from wiliot_core import set_logger, GetApiKey
from wiliot_tools.resolver_tool.resolve_packets import ResolvePackets
from wiliot_tools.utils.wiliot_gui.wiliot_gui import WiliotGui, popup_message
from wiliot_testers.utils.get_version import get_version
from wiliot_testers.tester_utils import dict_to_csv
from wiliot_testers.utils.upload_to_cloud_api import upload_to_cloud_api
from wiliot_testers.yield_tester.modules.first_gui import open_session, preparing_layout, ConversionTypes
from wiliot_testers.yield_tester.modules.adva_process import AdvaProcess
from wiliot_testers.yield_tester.modules.sensors_yield import SensorsYield
from wiliot_testers.yield_tester.modules.yield_gui import YieldPlotting
from wiliot_testers.yield_tester.utils.resolve_utils import ENV_RESOLVE, YieldTagStatus


DEFAULT_USER_INPUTS = {
    "max_temperature": 40,
    "min_temperature": 10,
    "temperature_type": "C",
    "min_humidity": 30,
    "max_humidity": 60,
    "min_light_intensity": 0,
    "max_light_intensity": 100,
    "max_y_axis": 110,
    "red_line_cumulative": 90,
    "red_line_current": 70,
    "pin_number": "004",
    "rssi_threshold": 80,
    "time_between_matrices_sec": 5,
    "resolve_wait_after_run": "no",
    "ignore_adva_before_triggers": True,
    "port": 8008,
    "host": "127.0.0.1",
    "n_matrix_to_estimate_wafer": 5
        }


matplotlib.use('TkAgg')


class MainWindow:
    """
    The main class the runs the GUI and supervise the multi-threading process of fraction's calculation and GUI viewing
    """

    def __init__(self, do_init_app_config=True):
        self.tester_type = 'yield'
        self.main_gui_instance = None
        self.first_gui_vals = None
        self.last_processed_index = 0
        self.current_values = None
        self.main_gui = None
        self.test_started = True
        self.logger = None
        self.adva_process = None
        self.adva_process_thread = None
        self.resolver_thread = None
        self.yield_gui_thread = None
        self.sensors = None
        self.sensors_thread = None
        self.resolve_path = ''
        self.external_ids = {}
        self.estimated_lot_wafer_pair = None
        self.n_ignore_external_ids = 0
        self.resolver = None
        self.resolve_q = Queue(maxsize=1000)
        self.packet_and_triggers_q = Queue(maxsize=1000)
        self.user_event_q = Queue(maxsize=100)
        self.folder_path = None
        self.run_data_path = None
        self.run_data_dict = None
        self.stop = threading.Event()
        self.cmn = ''
        self.packets_data_path = None
        self.advas_before_tags = set()
        self.fig_canvas_agg1 = None
        self.machine_type = 'assembly_yield_tester'
        self.yield_df = pd.DataFrame({'matrix_advas': 0, 'matrix_external_ids': 0, 'trigger_time': 0.0}, index=[0])
        self.user_inputs = self.init_user_inputs()
        self.inlays = self.init_inlays()

        if do_init_app_config:
            self.init_app_configuration()

    def init_app_configuration(self):
        # pop up GUI configuration:
        self.first_gui_setup()

        # init all proccess based on user selection
        self.setup_logger_and_paths()
        self.init_processes()
        self.init_run_data()
        self.start_processes()

    ###############  BEFORE RUN FUNCTIONS ###############

    def setup_logger_and_paths(self):
        """
        Sets logger and paths for the run
        @return:
        """
        cur_time_formatted = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
        self.cmn = self.get_cmn(cur_time_formatted)
        logger_path, self.logger = set_logger(app_name=self.machine_type, common_run_name=self.cmn,
                                              folder_name=self.get_folder_name())
        self.folder_path = os.path.dirname(logger_path)
        self.run_data_path = os.path.join(self.folder_path, self.cmn + '@run_data.csv')
        self.packets_data_path = os.path.join(self.folder_path, self.cmn + '@packets_data.csv')
        self.resolve_path = os.path.join(self.folder_path, self.cmn + '@resolve_data.csv')

    def init_inlays(self):
        inlay_data_path = os.path.join(os.path.dirname(__file__), 'configs', 'inlay_data.json')
        inlay_data_eng_path = os.path.join(os.path.dirname(__file__), 'configs', 'inlay_data_eng.json')
        if os.path.exists(inlay_data_eng_path):
            with open(inlay_data_eng_path) as f:
                inlays = json.load(f)
        else:
            with open(inlay_data_path) as f:
                inlays = json.load(f)
        return inlays

    def init_user_inputs(self, default_values=DEFAULT_USER_INPUTS): 
        script_dir = os.path.dirname(__file__)
        json_file_path = os.path.join(script_dir, 'configs', 'user_inputs.json')
        try:
            with open(json_file_path) as f:
                user_inputs = json.load(f)
            for key, value in default_values.items():
                if key not in user_inputs:
                    user_inputs[key] = value
            with open(json_file_path, 'w') as f:
                json.dump(user_inputs, f, indent=4)
        except Exception as e:
            print('could not read the user input file, default values are set')
            user_inputs = default_values
            os.makedirs(os.path.dirname(json_file_path), exist_ok=True)
            with open(json_file_path, 'w') as f:
                json.dump(user_inputs, f, indent=4)
        return user_inputs

    def init_run_data(self):
        """
        Initialize run data csv file.
        @return:
        """
        py_wiliot_version = get_version()
        start_time = datetime.datetime.now()
        run_start_time = start_time.strftime("%Y-%m-%d %H:%M:%S")
        self.run_data_dict = {
            'common_run_name': self.cmn, 'tester_station_name': self.first_gui_vals.get('tester_station_name', ''),
            'gw_version': self.adva_process.get_gw_version(),
            'tester_type': self.tester_type,
            'operator': self.first_gui_vals.get('operator', ''), 
            'received_channel': self.inlays[self.first_gui_vals['selected']]['received_channel'],
            'run_start_time': run_start_time, 'run_end_time': '',
            'wafer_lot': self.first_gui_vals.get('wafer_lot', ''),
            'wafer_number': self.first_gui_vals.get('wafer_number', ''),
            'assembled reel': self.first_gui_vals.get('assembled_reel', ''),
            'lane_ids': self.first_gui_vals.get('lane_ids', ''),
            'window_size': self.first_gui_vals.get('window_size', ''), 'upload_date': '',
            'gw_energy_pattern': self.first_gui_vals.get('gw_energy_pattern', ''),
            'comments': self.first_gui_vals.get('comments', ''), 'inlay': self.first_gui_vals['selected'], 'total_run_tested': 0,
            'total_run_responding_tags': 0,
            'conversion_type': ConversionTypes.NOT_CONVERTED.value,
            'surface': self.first_gui_vals.get('surface', ''),
            'matrix_tags': str(self.first_gui_vals.get('matrix_size', '')), 'py_wiliot_version': py_wiliot_version,
            'number_of_columns': self.first_gui_vals.get('thermodes_col', ''),
            'number_of_lanes': self.first_gui_vals.get('rows_number', ''),
            'gw_time_profile': self.first_gui_vals.get('gw_time_profile', '')
        }

    def first_gui_setup(self):
        """
        Does all the work of the first gui.
        @return:
        """
        previous_input, inlay_info_dict = preparing_layout(inlays=self.inlays)
        layout, cols_or_rows = self.open_session_layout(previous_input=previous_input, inlay_info=inlay_info_dict)
        self.first_gui_vals = open_session(open_session_layout=layout, cols_or_rows=cols_or_rows, inlays=self.inlays, 
                                           gui_title=self.machine_type.replace('_', ' ').title())

    def get_cmn(self, time_str):
        """
        Creates the common run name of the files we want to save.
        @param day_str: Date of the day
        @param time_str: time stamp
        @return: common run name
        """
        return '{wafer_lot}.{wafer_number}_{time}'.format(wafer_lot=self.first_gui_vals['wafer_lot'],
                                                                wafer_number=self.first_gui_vals['wafer_number'],
                                                                time=time_str)

    def get_folder_name(self):
        return '{}.{}'.format(self.first_gui_vals['wafer_lot'], self.first_gui_vals['wafer_number'])

    def open_session_layout(self, previous_input, inlay_info):
        """
        Returns GUI as dictionary for wiliotGUI
        @param previous_input: Dictionary of default values.
        @param inlay_info: Info about Inlay we take from data_inlay/data_inlay_eng
        @return: GUI as dictionary for wiliotGUI
        """
        cols_or_rows = 'thermodes_col'
        open_session_layout = {
            'wafer_lot': {'text': 'Wafer Lot:', 'value': previous_input['wafer_lot'], 'widget_type': 'entry'},
            'wafer_num': {'text': 'Wafer Number:', 'value': previous_input['wafer_num'], 'widget_type': 'entry'},
            'thermodes_col': {'text': 'Number of Columns:', 'value': previous_input['thermodes_col'],
                              'widget_type': 'entry'},
            'matrix_tags': {'text': '', 'widget_type': 'label',
                            'value': f'Total Tags per Matrix: {str(inlay_info["default_matrix_tags"])}'},
            'inlay_dict': [
                {'inlay': {'text': 'Inlay:', 'value': previous_input['inlay_dict_inlay'], 'widget_type': 'combobox',
                           'options': list(self.inlays.keys())}},
                {'inlay_info': {'widget_type': 'label', 'value': inlay_info['inlay_info']}},
            ],
            'tester_station_name': {'text': 'Tester Station:', 'value': previous_input['tester_station_name'],
                                    'widget_type': 'entry'},
            'comments': {'text': 'Comments:', 'value': previous_input['comments'], 'widget_type': 'entry'},
            'operator': {'text': 'Operator:', 'value': previous_input['operator'], 'widget_type': 'entry'},
            'surface': {'text': 'Surface:', 'value': previous_input['surface'], 'widget_type': 'combobox',
                        'options': inlay_info["surfaces"]},
            'window_size': {'text': 'Window Size for Analysis:', 'value': previous_input['window_size'],
                            'widget_type': 'entry'},
            'do_resolve': {'text': 'Get External Id from Cloud', 'value': previous_input['do_resolve']},
            'owner_id': {'text': 'Owner Id for Cloud Connection', 'value': previous_input['owner_id']},

        }

        return open_session_layout, cols_or_rows

    def get_app_errors(self):
        is_error = ''
        if self.adva_process is not None:
            if self.adva_process.gw_error_connection:
                is_error += (' Gateway was disconnected! Please try to manually disconnect and reconnect.')
        if self.estimated_lot_wafer_pair is not None:
            lot = self.estimated_lot_wafer_pair['lot']
            wafer = self.estimated_lot_wafer_pair['wafer']
            if lot != self.first_gui_vals.get('wafer_lot') or str(wafer).zfill(2) != str(self.first_gui_vals.get('wafer_number', 'N/A')).zfill(2):
                is_error += (f' Estimated Lot-{lot} Wafer-{wafer} is DIFFERENT than the user input.')
        return is_error
        

    ###############  RUNNING CLASS FUNCTIONS ###############

    def init_processes(self, sensors_type=None):
        """
        Initializing the two main instances and threads in order to start working
        @param inlay_select: Inlay type we are running
        @param rssi_th: RSSI threshold we want to filter according to it.
        @return:
        """
        state=''
        try:
            # init resolve
            state = 'init resolve'
            if self.first_gui_vals['do_resolve']:
                self.resolver = ResolvePackets(tags_in_test=[], owner_id=self.first_gui_vals['owner_id'], env=ENV_RESOLVE, resolve_q=self.resolve_q,
                                               set_tags_status_df=self.updated_resolved_tags, stop_event_trig=self.stop,
                                               logger_name=self.logger.name, gui_type='ttk', tag_status=YieldTagStatus,
                                               wait_after_run=self.user_inputs.get('resolve_wait_after_run').lower() == 'yes',
                                               do_parallel_request=True)
                self.resolver_thread = threading.Thread(target=self.resolver.run, args=())
            # init sensors:
            state = 'init sensors'
            self.sensors = SensorsYield(stop_event=self.stop, logger=self.logger, sensors_type=sensors_type)
            if self.sensors.is_sensor_enable():
                self.sensors_thread = threading.Thread(target=self.sensors.run, args=())
            # init adva process:
            state = 'init adva process'
            adva_process_inputs = {'selected_inlay': self.inlays.get(self.first_gui_vals['selected']), 'listener_path': self.folder_path,
                                   'cmn': self.cmn, 'user_inputs': self.user_inputs}
            self.adva_process = AdvaProcess(stop_event=self.stop, logger=self.logger,
                                            adva_process_inputs=adva_process_inputs,
                                            packet_and_triggers_q=self.packet_and_triggers_q,
                                            get_sensors_values=self.sensors.get_sensors_data,
                                            ignore_adva_before_triggers=self.user_inputs['ignore_adva_before_triggers'])
            self.adva_process_thread = threading.Thread(target=self.adva_process.run, args=())

            #init GUI process:
            state = 'init GUI process'
            self.yield_gui = YieldPlotting(get_data=self.get_yield_df,
                                           get_sensors_data=self.sensors.get_sensors_data,
                                           get_app_errors=self.get_app_errors,
                                           cmd_q=self.user_event_q,
                                           user_config={**self.first_gui_vals, **self.user_inputs},
                                           logger=self.logger,
                                           stop_event=self.stop,
                                           tester_name=self.machine_type)
            self.yield_gui_thread = threading.Thread(target=self.yield_gui.run, args=())
            webbrowser.open(f"http://{self.user_inputs['host']}:{self.user_inputs['port']}/")
            
        except Exception as e:
            self.logger.warning(f"{e}")
            popup_message(msg=f'{state} FAILED due to:\n{e}', logger=self.logger)
            sys.exit(-1)

    def start_processes(self):
        """
        Starting the work of the both threads
        @return:
        """
        self.adva_process_thread.start()
        if self.sensors_thread is not None:
            self.sensors_thread.start()
        if self.resolver_thread is not None:
            self.resolver_thread.start()
        self.yield_gui_thread.start()

    def run(self):
        """
        Viewing the window and checking if the process stops
        @return:
        """
        self.logger.info('Start Yield Update Data Run')
        while True:
            try:
                if self.stop.is_set():
                    self.logger.info('Stop Yield Update Data Run')
                    break

                self.update_data()
                self.estimate_last_matrices_wafer()
                self.handle_user_events()
                time.sleep(0.100 if self.packet_and_triggers_q.empty() else 0)
            except Exception as e:
                self.logger.warning(f'got exception during yield update data run: {e}')

        self.stop_yield()

    def handle_user_events(self):
        if self.user_event_q.empty():
            return
        user_event = self.user_event_q.get()
        if 'start' in user_event.lower():
            self.adva_process.set_stopped_by_user(stopped=False)
        elif 'pause' in user_event.lower():
            self.adva_process.set_stopped_by_user(stopped=True)
        elif 'stop' in user_event.lower():
            self.stop.set()
        else:
            self.logger.warning(f'handle_user_events: Unsupported commamd: {user_event}')


    def stop_yield(self):
        """
        All the work when the application is stopped (User interaction or Error).
        @return:
        """
        final_tags = self.get_number_of_tested()
        total_ex_ids = self.yield_df['matrix_external_ids'].sum()
        total_advas = self.yield_df['matrix_advas'].sum()

        self.logger.info(
            f'Final Adva Yield: {(total_advas / final_tags) * 100 if final_tags > 0 else 0}%, '
            f'Final External Ids Yield: {(total_ex_ids / final_tags if final_tags > 0 else 0) * 100}%, '
            f'Final Tags: {final_tags}, Final Advas: {total_advas}, Final External Ids: {total_ex_ids}'
            )
        self.logger.info(f"User quit from application")

        # wait to all thread to join
        threads = {'adva process': self.adva_process_thread, 'resolver': self.resolver_thread, 'sensors': self.sensors_thread, 'gui': self.yield_gui_thread}
        for n, thr in threads.items():
            if thr is None:
                continue
            thr.join(timeout=5)
            if thr.is_alive():
                self.logger.warning(f'thread {n} is still running')
            else:
                self.logger.info(f'thread {n} is completed')
        
        self.update_data()

        is_uploaded = self.upload_to_cloud()
        self.update_run_data_file(is_uploaded)
        sys.exit()

    ###############  RESOLVE FUNCTIONS ###############

    def updated_resolved_tags(self, tag_status):
        """
        Updating resolve data.
        """
        tag_status = {k: v[0] for k, v in tag_status.items()}
        new_ex_id = tag_status['external_id']
        if new_ex_id.lower() in ['unknown', 'n/a']:
            new_ex_id += tag_status['tag']
        matrix_num = tag_status['matrix_num']
        self.logger.info(f'update resolved tags: {new_ex_id}')
        if new_ex_id not in self.external_ids.keys():
            self.external_ids[new_ex_id] = matrix_num
            # filter based on the lot/wafer
            estimated_lot_wafer = self.extract_lot_wafer_from_ex_id(ex_id=new_ex_id)
            added_ex_id = False
            if estimated_lot_wafer != []:
                lot, wafer = estimated_lot_wafer[0]
                if lot == self.first_gui_vals.get('wafer_lot') and str(wafer).zfill(2) == str(self.first_gui_vals.get('wafer_number', 'N/A')).zfill(2):
                    self.yield_df.loc[matrix_num, 'matrix_external_ids'] += 1
                    added_ex_id = True
            if not added_ex_id:
                self.n_ignore_external_ids += 1
                self.logger.info(f'filter out ex_id: {new_ex_id} from external id count. Currently {self.n_ignore_external_ids} tags were filtered-out')
        
        dict_to_csv(dict_in=tag_status, path=self.resolve_path, append=os.path.exists(self.resolve_path))

    def add_to_resolve_queue(self, packet_list_in, adva_to_matrix_location):
        """
        Adds resolve data to the queue while running.
        """
        for packet_in in packet_list_in:
            adva = packet_in.get_adva()
            matrix_num = adva_to_matrix_location[adva_to_matrix_location['adv_address'] == adva]['matrix_tags_location']
            if matrix_num.empty:
                self.logger.warning(f'add_to_resolve_queue: could not find adva: {adva} in the adva_to_matrix_location df')
                continue
            matrix_num = int(matrix_num.values[-1])
            if self.resolve_q.full():
                self.logger.warning(f'Resolve queue is full. Discard the following adva: {adva}')
                continue
            self.resolve_q.put({'tag': adva, 'payload': packet_in.get_payload(), 'matrix_num': matrix_num})

    ###############  UPDATES FUNCTIONS ###############

    def process_new_df(self, new_data):
        new_df = new_data['packet'].get_df(sprinkler_filter=True)
                
        matrix_location_ind = np.searchsorted(self.yield_df['trigger_time'], new_df['time_from_start']) - 1

        new_df.insert(loc=0, column='common_run_name', value=self.cmn)
        new_df.insert(loc=1, column='matrix_tags_location', value=self.yield_df.index[matrix_location_ind].values)
        new_df.insert(loc=2, column='matrix_timestamp', value=self.yield_df['trigger_time'].iloc[matrix_location_ind].values)
        new_df.insert(loc=3, column='tag_matrix_ttfp', value=(new_df['time_from_start'] - new_df['matrix_timestamp']).values)
        
        new_df.insert(loc=len(new_df.columns), column='environment_light_intensity', value=new_data['sensors'].get('light_intensity'))
        new_df.insert(loc=len(new_df.columns), column='environment_humidity', value=new_data['sensors'].get('humidity'))
        new_df.insert(loc=len(new_df.columns), column='environment_temperature', value=new_data['sensors'].get('temperature'))
        return new_df

    def process_new_triggers(self, new_triggers):
        new_rows_df = pd.DataFrame()
        new_rows = {'trigger_num': [], 'trigger_time': []}
        for t in new_triggers:
            if t['trigger_num'] in self.yield_df.index:
                continue
            new_rows['trigger_num'].append(t['trigger_num'])
            new_rows['trigger_time'].append(t['trigger_time'])
        if new_rows['trigger_num']:
            new_rows_df = pd.DataFrame({'trigger_time': new_rows['trigger_time']}, index=new_rows['trigger_num'])
            new_rows_df.insert(loc=len(new_rows_df.columns), column='matrix_advas', value=0)
            new_rows_df.insert(loc=len(new_rows_df.columns), column='matrix_external_ids', value=0)

        return new_rows_df


    def update_data(self):
        """
        Updates the run_data CSV file while running the program
        @return:
        """
        if self.packet_and_triggers_q.empty():
            return
        
        new_data = self.packet_and_triggers_q.get()
        # update triggers
        new_matrix_df = self.process_new_triggers(new_data['trigger'])
        self.yield_df = pd.concat([self.yield_df, new_matrix_df])

        if len(new_data['packet']):
            # update main df
            new_df = self.process_new_df(new_data=new_data)
            new_n_advas = new_df.groupby('matrix_tags_location').size()
            self.yield_df.loc[new_n_advas.index, 'matrix_advas'] += new_n_advas.values

            # send to resolve:
            if self.first_gui_vals['do_resolve']:
                self.add_to_resolve_queue(new_data['packet'], new_df)

            # log files
            if not os.path.exists(self.packets_data_path):
                new_df.to_csv(self.packets_data_path, mode='w', header=True, index=False)
            else:
                new_df.to_csv(self.packets_data_path, mode='a', header=False, index=False)
        
        self.update_run_data_file()
        self.logger.info(f"last 10 rows of yield df:\n{self.yield_df.tail(10)}")

    def update_run_data_file(self, is_uploaded=False):

        """
        Updates the run_data CSV file while running the program
        @param is_uploaded: If to update the upload date at the end of the run or not
        @return:
        """
        end_time = datetime.datetime.now()
        run_end_time = end_time.strftime("%Y-%m-%d %H:%M:%S")
        advas = self.yield_df['matrix_advas'].sum()
        tags_num = self.get_number_of_tested()
        result = float(100 * (advas / tags_num)) if tags_num != 0 else float('nan')

        if is_uploaded:
            self.run_data_dict['upload_date'] = run_end_time
        self.run_data_dict['run_end_time'] = run_end_time
        self.run_data_dict['total_run_tested'] = tags_num
        self.run_data_dict['total_run_responding_tags'] = advas
        self.run_data_dict['yield'] = result
        if self.first_gui_vals['do_resolve']:
            self.run_data_dict['total_run_external_ids'] = self.yield_df['matrix_external_ids'].sum()
            self.run_data_dict['yield_external_ids'] = 100 * (
                    self.run_data_dict['total_run_external_ids'] / tags_num) if tags_num else 0
        dict_to_csv(dict_in=self.run_data_dict, path=self.run_data_path)

    def get_number_of_tested(self):
        """
        Return number of tested tags.
        @return: number of tested tags.
        """
        tags_num = self.yield_df.index[-1] * int(self.first_gui_vals['matrix_size'])
        return tags_num
    
    def get_yield_df(self):
        return self.yield_df
    
    @staticmethod
    def extract_lot_wafer_from_ex_id(ex_id, lot_wafer_pairs=None):
        lot_wafer_pairs = lot_wafer_pairs if lot_wafer_pairs is not None else []
        pattern = r"Lot-(\w+)__Wafer-(\w+)__X-.*__Y-.*"
        match = re.search(pattern, ex_id)
        if match:
            lot, wafer = match.group(1), match.group(2)
            lot_wafer_pairs.append((lot, wafer))
        return lot_wafer_pairs

    def estimate_last_matrices_wafer(self):
        n_matrix_to_estimate = self.user_inputs.get('n_matrix_to_estimate_wafer', 5)
        max_index = self.yield_df.index.max()
        last_estimated_index = self.estimated_lot_wafer_pair.get('last_matrix_index', 0) if self.estimated_lot_wafer_pair else 0
        if max_index - last_estimated_index < n_matrix_to_estimate:
            return
        # estimated wafer based on the resolved tags: external_id = f"Lot-{lot}__Wafer-{row.Wafer}__X-{row.X}__Y-{row.Y}"
        if len(self.external_ids) == 0:
            return
        # Extract all lot-wafer pairs on the last n_matrix_to_estimate
        min_matrix_index = max_index - n_matrix_to_estimate
        lot_wafer_pairs = []
        for ex_id, matrix_num in self.external_ids.items():
            if matrix_num > 0 and matrix_num >= min_matrix_index:
                lot_wafer_pairs = self.extract_lot_wafer_from_ex_id(ex_id, lot_wafer_pairs)
        
        # Find most common pair
        self.logger.info(f'Got the following lot-wafer based on last {n_matrix_to_estimate}: {set(lot_wafer_pairs)}')
        if lot_wafer_pairs:
            most_common = Counter(lot_wafer_pairs).most_common(1)[0]
            self.logger.info(f'Estimated lot-wafer: Lot-{most_common[0][0]} Wafer-{most_common[0][1]} with count {most_common[1]}/{len(lot_wafer_pairs)}')
            self.estimated_lot_wafer_pair = {'lot': most_common[0][0], 'wafer': most_common[0][1], 'last_matrix_index': max_index}
        else:
            self.estimated_lot_wafer_pair = {'lot': 'N/A', 'wafer': 'N/A', 'last_matrix_index': max_index}
        return

    ################ UPLOAD FUNCTION ###############
    def do_upload(self, env_choice, owner_id):
        try:
            # make sure user have the api key:
            GetApiKey(owner_id=owner_id, env=env_choice)
            tester_type = self.tester_type.replace('_', '-') + '-test'
            is_uploaded = upload_to_cloud_api(batch_name=self.cmn, tester_type=tester_type,
                                              run_data_csv_name=self.run_data_path, env=env_choice, is_path=True,
                                              packets_data_csv_name=self.packets_data_path, owner_id=owner_id)

        except Exception as ee:
            is_uploaded = False
            self.logger.warning(f"do_upload: Exception occurred: {ee}")
        return is_uploaded

    def upload_to_cloud(self):
        """
        All the process of uploading data to cloud.
        @return:
        """
        is_uploaded = False
        yes_or_no = ['Yes', 'No']
        upload_layout = {'ask_to_upload': {'widget_type': 'label', 'value': 'Do you want to stop or upload?'},
                             'upload': {'text': 'Upload:', 'value': yes_or_no[0], 'widget_type': 'combobox',
                                        'options': yes_or_no},
                             'env_choice': {'text': 'Select Environment:', 'value': 'prod', 'widget_type': 'combobox',
                                            'options': ['prod', 'test']},
                             'owner_id': {'text': 'Owner Id', 'value': self.first_gui_vals['owner_id'],
                                          'widget_type': 'entry'}
                             }
        upload_gui = WiliotGui(params_dict=upload_layout, title='Upload to cloud', exit_sys_upon_cancel=False)
        values_out = upload_gui.run()
        if values_out and values_out['upload'] == 'Yes':
            is_uploaded = self.do_upload(env_choice=values_out['env_choice'], owner_id=values_out['owner_id'])
            if is_uploaded:
                self.logger.info("Successful upload")
                popup_message(msg="Successfully upload data to cloud!", logger=self.logger, bg='green')
            else:
                self.logger.info('Failed to upload the file')
                popup_message(msg="Run upload failed.\n Please check Internet connection and upload logs manually", logger=self.logger, bg='red')

        else:
            self.logger.info('File was not uploaded')
        return is_uploaded
if __name__ == '__main__':
    m = MainWindow()
    m.run()



from dash import Dash, html, dcc, Input, Output, State, clientside_callback, callback_context
import dash_bootstrap_components as dbc
import os
import pandas as pd
import plotly.graph_objects as go
import traceback

from wiliot_core.utils.utils import open_directory
from wiliot_testers.failure_analysis_tester.fa_dash_components import *
from wiliot_testers.failure_analysis_tester.fa_utils import *
from wiliot_testers.failure_analysis_tester.configs_gui import CONFIG_FILE, CONFIGS_DIR, save_config, load_config
from wiliot_testers.failure_analysis_tester.failure_analysis_tester import FailureAnalysisTester

ELEMENT_WIDTH = "300px"
BUTTON_WIDTH = 30
TEST_OPTIONS = [
    {'label': 'IV Curve', 'value': 'IV Curve'},
    {'label': 'Voltage Drop', 'value': 'Voltage Drop'},
    {'label': 'External Power Source', 'value': 'External Power Source'}
]

MAIN_TITLE = html.H3(["Failure Analysis Tester"], className="bg-info p-3 text-center")
DISABLED = False
KILL_APP = False
config = load_config()

try:
    FAT = FailureAnalysisTester()
except ConnectionError as e:
    MAIN_TITLE = html.H3(["Arduino is not connected/has an error.\nPlease close all windows, fix issue and run again"], className="bg-danger p-3 text-center")
    DISABLED = True
    KILL_APP = True
except ValueError as e:
    MAIN_TITLE = html.H3(["SMU is not connected/has an error.\nPlease make sure SMU address is correct"], className="bg-danger p-3 text-center")
    DISABLED = True
except Exception as e:
    MAIN_TITLE = html.H3([f"Unknown error occurred. Please contact Wiliot.\n Error traceback {traceback.format_exc()}\n Error message {str(e)}"], className="bg-danger p-3 text-center")
    DISABLED = True
    KILL_APP = True

app = Dash(__name__, external_stylesheets=[
           dbc.themes.MINTY, dbc.icons.FONT_AWESOME])

color_mode_switch = html.Span(
    [
        dbc.Label(className="fa fa-moon", html_for="color-mode-switch"),
        dbc.Switch(id="color-mode-switch", value=False,
                   className="d-inline-block ms-1", persistence=True),
        dbc.Label(className="fa fa-sun", html_for="color-mode-switch"),
    ]
)

clientside_callback(
    """
    (switchOn) => {
       document.documentElement.setAttribute('data-bs-theme', switchOn ? 'light' : 'dark');
       return window.dash_clientside.no_update
    }
    """,
    Output("color-mode-switch", "id"),
    Input("color-mode-switch", "value"),
)

app.layout = dbc.Container(
    [
        dcc.ConfirmDialog(
            id="confirm", message="Are you sure you want to end the test?"),
        color_mode_switch,
        html.Div(id='title-container', children=[MAIN_TITLE]),
        html.Br(),
        dbc.Modal(
            [
                dbc.ModalBody("Folder Run Name and Tag Alias are required to run the test"),
                dbc.ModalFooter(dbc.Button("Close", id="close_modal", className="ms-auto"))
            ],
            id="modal",
            is_open=False,  # start closed
            backdrop=True,  # click outside to close
            keyboard=True,  # ESC to close
            centered=True,
        ),
        html.Div([
            get_tag_alias(disabled=DISABLED),
            get_folder_name(disabled=DISABLED),
            get_comment(disabled=DISABLED),
            get_test_and_tag_type(disabled=DISABLED),
            get_run_finish_buttons(disabled=DISABLED),
            get_test_checkbox(),
            get_test_result(disabled=DISABLED),
            get_smu_config(default_value=config['visa_addr']),
        ], style={'display': 'grid', 'gridTemplateColumns': 'repeat(4, 1fr)', 'gap': '10px', }),
        html.Br(),
        dcc.Loading(id="loading-output", type="default",
                    children=[html.Div(id="output")]),
        html.Br(),
        get_plot_selection(disabled=DISABLED),
        html.Br(),
        dcc.Graph(id='graph-content', style={"display": "none"}),
        dcc.Store(id="kill_trigger"),  # hidden target

    ],
)


@app.callback(
    Output('graph-content', 'figure'),
    Output('graph-content', 'style'),
    Input('plot-selection', 'value'),
    prevent_initial_call=True
)
def update_graph(value):
    if value == 'Empty':
        return go.Figure(), {"display": "none"}
    if FAT.is_app_running:
        df = FAT.df
        tag_alias = df['tag_alias'].unique()[0]
    else:
        tag_alias = ''
        df = FAT.all_df

    if  value + '_current_uA' not in df.columns or value + '_voltage_V' not in df.columns:
        return go.Figure(), {"display": "none"}
    
    fig = plot_value(df, FAT.reference_df, value, tag_alias)

    return fig, {"display": "block", "height":"700px"}


@app.callback(
    Output("smu_address_button", "n_clicks"),
    Input("smu_address_button", "n_clicks"),
    State("smu_address", "value"),
    prevent_initial_call=True
)
def set_smu_address(n_clicks, smu_address):
    if n_clicks:
        config['visa_addr'] = smu_address
        save_config(config)
        if DISABLED:
            os._exit(0)
    return 0


@app.callback(
    Output("kill_trigger", "data"),
    Input("finish_test", "disabled"),
    prevent_initial_call=True
)
def kill_app(disabled):
    if disabled and not FAT.is_app_running:
        os._exit(0)
    return disabled
    

@app.callback(
    Output("confirm", "displayed", allow_duplicate=True),
    Output("run_test", "disabled", allow_duplicate=True),
    Output("test_comment", "disabled", allow_duplicate=True),
    Output("test_type", "disabled", allow_duplicate=True),
    Output("tag_alias", "disabled", allow_duplicate=True),
    Output("finish_test", "disabled", allow_duplicate=True),
    Output("tag_type", "disabled", allow_duplicate=True),
    Output("folder_name", "disabled", allow_duplicate=True),
    Output('title-container', 'children'),
    Input("confirm", "submit_n_clicks"),
    State("folder_name", "disabled"),
    State("tag_type", "disabled"),
    prevent_initial_call=True
)
def close_app(n_clicks, folder_name_disabled, tag_type_disabled):
    title = [MAIN_TITLE]
    if n_clicks:
        FAT.is_app_running = False
        FAT.end_test()
        title = [html.H3(["Failure Analysis Tester run is done, to run again please close all windows"], className="bg-danger p-3 text-center")]
        return False, True, True, True, True, True, True, True, title
    return False, False, False, False, False, False, tag_type_disabled, folder_name_disabled, title


@app.callback(
    Output("confirm", "displayed"),
    Input("finish_test", "n_clicks"),
    prevent_initial_call=True
)
def show_confirm(n_clicks):
    return True  # Triggers the pop-up


@app.callback(
    Output("test_result", "style"),
    Input("test_result", "value"),
    prevent_initial_call=True
)
def update_input_style(value):
    color_map = {
        "Fail": "#ffcccc",
        "Pass": "#28a745"
    }
    return {"backgroundColor": color_map.get(value, "")}


@app.callback(
    Output("test_fields", "options"),
    Input("test_type", "value"),
    State("test_fields", "options"),
    prevent_initial_call=True
)
def set_test_fields_disabled(test_type, test_options):
    if test_type == 'IV Curve':
        for option in test_options:
            option['disabled'] = False
    else:
        for option in test_options:
            option['disabled'] = True
    return test_options
    

@app.callback(
    Output("run_test", "disabled", allow_duplicate=True),
    Output("test_comment", "disabled", allow_duplicate=True),
    Output("test_type", "disabled", allow_duplicate=True),
    Output("tag_type", "disabled", allow_duplicate=True),
    Output("tag_alias", "disabled", allow_duplicate=True),
    Output("finish_test", "disabled", allow_duplicate=True),
    Output("open_output_folder", "disabled"),
    Output("test_result", "value"),
    Output("loading-output", "children"),
    Output("folder_name", "disabled", allow_duplicate=True),
    Output("modal", "is_open", allow_duplicate=True),
    Input("run_test", "disabled"),
    State("test_type", "value"),
    State("tag_type", "value"),
    State("tag_type", "disabled"),
    State("folder_name", "value"),
    State("tag_alias", "value"),
    State("test_comment", "value"),
    State("test_fields", "value"),
    State("open_output_folder", "disabled"),
    prevent_initial_call=True
)
def run_test(disabled, test_type, tag_type, tag_type_disabled, folder_name, tag_alias, comment, test_fields, open_output_folder_disabled):
    test_result = ''
    if not tag_alias or not folder_name:
        return False, False, False, tag_type_disabled, False, False, open_output_folder_disabled, test_result, None, folder_name is not None, True
    if disabled and FAT.is_app_running:
        FAT.run_test(test_type=test_type, folder_name=folder_name.strip(),
                     tag_alias=tag_alias, comment=comment, keys=test_fields)
        test_result = FAT.check_test(tag_type, test_type)
        FAT.fill_list_for_report(test_type=test_type, tag_alias=tag_alias)
        return False, False, False, tag_type_disabled, False, False, False, test_result, None, True, False
    else:
        return True, True, True, True, True, True, False, '', None, True, False


@app.callback(
    Output("folder_name", "disabled", allow_duplicate=True),
    Output("test_comment", "disabled"),
    Output("test_type", "disabled"),
    Output("tag_type", "disabled"),
    Output("tag_alias", "disabled"),
    Output("run_test", "disabled"),
    Output("finish_test", "disabled"),
    Output('plot-selection', 'value'),
    Output('test_result', 'value', allow_duplicate=True),
    Input("run_test", "n_clicks"),
    prevent_initial_call=True
)
def run_test_button(n_clicks):
    return True, True, True, True, True, True, True, 'Empty', ''

@app.callback(
    Output("modal", "is_open", allow_duplicate=True),
    Input("close_modal", "n_clicks"),
    prevent_initial_call=True
)
def toggle_modal(n_clicks):
    return False

@app.callback(
    Output("open_output_folder", "n_clicks"),
    Input("open_output_folder", "n_clicks"),
    prevent_initial_call=True
)
def open_output_folder(n_clicks):
    if n_clicks:
        if FAT.output_dir:
            open_directory(FAT.output_dir)
    return 0


if __name__ == "__main__":
    import webbrowser
    import threading
    import time

    def exit_func():
        time.sleep(5)
        os._exit(0)

    webbrowser.open("http://127.0.0.1:8050/")
    if KILL_APP:
        exit_thread = threading.Thread(target=exit_func, args=())
        exit_thread.start()
    app.run_server()
